import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter
from ablelabs.neon.common.notable.nanophilia.structs import Location, Speed, FlowRate


class MotionAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.motion_initialize)
    async def initialize(self):
        pass

    @run_server_func(RobotRouter.motion_initialize_for_washer_dryer)
    async def initialize_for_washer_dryer(self):
        pass

    @run_server_func(RobotRouter.motion_move_z_up)
    async def move_z_up(self):
        pass

    @run_server_func(RobotRouter.motion_move_to_ready)
    async def move_to_ready(self):
        pass

    @run_server_func(RobotRouter.motion_move_to)
    async def move_to(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_pick_up_tip)
    async def pick_up_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_drop_tip)
    async def drop_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_rise_tip)
    async def rise_tip(
        self,
        pipette_number: int,
        height_offset: float,
        z_speed: Speed,
    ):
        pass

    @run_server_func(RobotRouter.motion_ready_plunger)
    async def ready_plunger(
        self,
        pipette_number: int,
        flow_rate: FlowRate = None,
    ):
        pass

    @run_server_func(RobotRouter.motion_blow_out)
    async def blow_out(
        self,
        pipette_number: int,
        flow_rate: FlowRate = None,
    ):
        pass

    @run_server_func(RobotRouter.motion_aspirate)
    async def aspirate(
        self,
        pipette_number: int,
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_dispense)
    async def dispense(
        self,
        pipette_number: int,
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_mix)
    async def mix(
        self,
        pipette_number: int,
        volume: float | list[float],
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate = None,
        delay: float = 0.0,
        optimize: bool = False,
    ):
        pass



# python 3.10 버전

# 가상환경 설정
## python -m venv .venv
## py 3.10 -m venv .venv

# ablelabs 패키지 설치
## pip install ablelabs

# Neon.exe + resources 파일 위치 확인
## resources 폴더는 Neon.exe와 같은 폴더 또는 하위 폴더에

# subprocess.Popen으로 Neon.exe 실행

# connect -> wait_boot 순서대로 호출


import asyncio
import sys, os
from typing import Literal

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.controllers.alma.api.robot_api import RobotAPI
from ablelabs.neon.common.alma.enums import (
    LocationType,
    Axis,
    LocationReference,
    RunStatus,
    ZGripper,
    DioInput,
)
from ablelabs.neon.common.alma.structs import (
    State,
    location,
    LCRParam,
    MeasurementTimeMode,
    HeaterData,
    Speed,
)


async def main():
    import subprocess

    # for windows
    # subprocess.Popen(
    #     [
    #         r"C:\your_path\Neon.exe",
    #     ]
    # )

    # for mac dev
    subprocess.Popen(
        [
            r"/Users/sypark/Code/ABLE-Elba/.venv/bin/python",
            r"/Users/sypark/Code/ABLE-Elba/robot/src/controllers/alma/robot_router.py",
        ],
        cwd=r"/Users/sypark/Code/ABLE-Elba",
    )
    await asyncio.sleep(1)

    ip = "localhost"
    port = 1234

    # robot
    robot_api = RobotAPI()
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass
    await robot_api.wait_boot()

    await robot_api.stop()
    await robot_api.clear_error()  # stop 이후, motion 전에. / fault 후에.
    await robot_api.pause()
    await robot_api.resume()
    is_connected: dict[
        Literal[
            "motor",
            "dio",
            "heater",
            "switcher",
            "lcr_meter",
        ],
        bool,
    ] = await robot_api.is_connected()
    environment: dict[
        Literal[
            "temperature",
            "pressure",
            "humidity",
        ],
        float,
    ] = await robot_api.get_environment()
    run_status: RunStatus = await robot_api.get_run_status()
    dio_input: dict[DioInput, bool] = await robot_api.get_dio_input()
    last_state: State = await robot_api.get_last_state()

    await robot_api.set_interlock(value=False)
    await robot_api.set_led_lamp(on=True)

    await robot_api.set_run_status(value=RunStatus.INITIALIZE)
    await robot_api.initialize(recovery=True)
    await robot_api.initialize_teaching()
    await robot_api.calibrate_lcr_meter()
    await asyncio.sleep(0.5)

    async def get_changed_robot_status():
        while True:
            robot_status = await robot_api.get_changed_robot_status()
            run_status: RunStatus = robot_status["run_status"]
            is_door_open: bool = robot_status["is_door_open"]
            dio_input: dict[DioInput, bool] = robot_status["dio_input"]

    asyncio.create_task(get_changed_robot_status())

    await robot_api.set_run_status(value=RunStatus.RUN)
    await robot_api.set_progress_rate(value=5)
    await robot_api.set_run_status(value=RunStatus.DONE)

    # get
    setup_data: dict = await robot_api.get.setup_data()

    # set
    x, y, z = setup_data["teaching"]["absolute_origin"]["pipette_deck_1"]
    setup_data["teaching"]["absolute_origin"]["pipette_deck_1"] = [x + 1, y + 2, z + 3]
    setup_data["option"]["pipette"]["cal_a"] = 1.321
    await robot_api.set.update_setup_data(value=setup_data, save=True)
    setup_data_2: dict = await robot_api.get.setup_data()
    await robot_api.set.update_pipette_attrs(
        value={
            2: {
                "blow_out_volume": 20,
            }
        }
    )

    # time
    sec = 0
    sec += await robot_api.time.initialize()
    sec += await robot_api.time.calibrate_lcr_meter()
    # time.upper_module
    sec += await robot_api.time.upper_module.move_z_up()
    sec += await robot_api.time.upper_module.move_to_ready()
    # time.upper_module.gripper
    sec += await robot_api.time.upper_module.gripper.move_block_to_inspector(push=True)
    sec += await robot_api.time.upper_module.gripper.move_block_to_holder()
    sec += await robot_api.time.upper_module.gripper.open_lid(
        chip_holder_number=1,
        lid_holder_number=2,
    )
    sec += await robot_api.time.upper_module.gripper.close_lid(
        lid_holder_number=2,
        chip_holder_number=1,
    )
    sec += await robot_api.time.upper_module.gripper.move_labware(
        from_location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=1,
        ),
        to_location=location(
            location_type=LocationType.INSPECTOR,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.move_lid(
        from_location=location(
            location_type=LocationType.LID_HOLDER,
            location_number=1,
        ),
        to_location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=2,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.pick_labware(
        location=location(
            location_type=LocationType.INSPECTOR,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.place_labware(
        location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=1,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.pick_lid(
        location=location(
            location_type=LocationType.LID_HOLDER,
            location_number=1,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.place_lid(
        location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=2,
        ),
    )
    sec += await robot_api.time.upper_module.gripper.move_to(
        location=location(
            location_type=LocationType.DECK,
            location_number=1,
        ),
        z_gripper=ZGripper.UNDER_PLATE,
        slow_z_offset=5,
    )
    # time.upper_module.pipette
    sec += await robot_api.time.upper_module.pipette.pick_up_tip(
        well="a1",
    )
    sec += await robot_api.time.upper_module.pipette.drop_tip(
        location_type=LocationType.WASTE,
        # location_type=LocationType.TIP_RACK,
        well="a1",
    )
    sec += await robot_api.time.upper_module.pipette.move_to(
        location=location(
            location_type=LocationType.RESERVOIR,
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(0, 0, 1.2),
        ),
    )
    sec += await robot_api.time.upper_module.pipette.mix(
        volume=[50, 50],
        iteration=3,
        flow_rate=100,
    )
    sec += await robot_api.time.upper_module.pipette.aspirate(
        volume=[50, 50],
        flow_rate=100,
    )
    wells = ["a1", "b1"]
    for i, well in enumerate(wells):
        sec += await robot_api.time.upper_module.pipette.move_to(
            location=location(
                location_type=LocationType.CHIP_HOLDER,
                location_number=1,
                well=well,
                reference=LocationReference.BOTTOM,
            ),
        )
        sec += await robot_api.time.upper_module.pipette.dispense(
            volume=50,
            height_offset=5,
        )
        is_last = i + 1 == len(wells)
        if is_last:
            sec += await robot_api.time.upper_module.pipette.blow_out()
        sec += await robot_api.time.upper_module.pipette.move_to(
            location=location(
                location_type=LocationType.CHIP_HOLDER,
                location_number=1,
                well=well,
                reference=LocationReference.TOP_JUST,
            ),
            z_speed=Speed.from_mm(10),
            optimize=True,
        )

    # upper_module
    await robot_api.upper_module.move_z_up()
    await robot_api.upper_module.move_to_ready()

    # upper_module.gripper
    await robot_api.upper_module.gripper.move_block_to_inspector(push=True)
    await robot_api.upper_module.gripper.move_block_to_holder()
    await robot_api.upper_module.gripper.open_lid(
        chip_holder_number=1,
        lid_holder_number=2,
    )
    await robot_api.upper_module.gripper.close_lid(
        lid_holder_number=2,
        chip_holder_number=1,
    )
    await robot_api.upper_module.gripper.move_labware(
        from_location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=1,
        ),
        to_location=location(
            location_type=LocationType.INSPECTOR,
        ),
    )
    await robot_api.upper_module.gripper.move_lid(
        from_location=location(
            location_type=LocationType.LID_HOLDER,
            location_number=1,
        ),
        to_location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=2,
        ),
    )
    await robot_api.upper_module.gripper.pick_labware(
        location=location(
            location_type=LocationType.INSPECTOR,
        ),
    )
    await robot_api.upper_module.gripper.place_labware(
        location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=1,
        ),
    )
    await robot_api.upper_module.gripper.pick_lid(
        location=location(
            location_type=LocationType.LID_HOLDER,
            location_number=1,
        ),
    )
    await robot_api.upper_module.gripper.place_lid(
        location=location(
            location_type=LocationType.CHIP_HOLDER,
            location_number=2,
        ),
    )
    await robot_api.upper_module.gripper.move_to(
        location=location(
            location_type=LocationType.DECK,
            location_number=1,
        ),
        z_gripper=ZGripper.UNDER_PLATE,
        slow_z_offset=5,
    )

    # upper_module.pipette
    await robot_api.upper_module.pipette.pick_up_tip(
        well="a1",
    )
    await robot_api.upper_module.pipette.drop_tip(
        location_type=LocationType.WASTE,
        # location_type=LocationType.TIP_RACK,
        well="a1",
    )
    await robot_api.upper_module.pipette.move_to(
        location=location(
            location_type=LocationType.RESERVOIR,
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(0, 0, 1.2),
        ),
    )
    await robot_api.upper_module.pipette.mix(
        volume=[50, 50],
        iteration=3,
        flow_rate=100,
    )
    await robot_api.upper_module.pipette.aspirate(
        volume=[50, 50],
        flow_rate=100,
    )
    wells = ["a1", "b1"]
    for i, well in enumerate(wells):
        await robot_api.upper_module.pipette.move_to(
            location=location(
                location_type=LocationType.CHIP_HOLDER,
                location_number=1,
                well=well,
                reference=LocationReference.BOTTOM,
            ),
        )
        await robot_api.upper_module.pipette.dispense(
            volume=50,
            height_offset=5,
        )
        is_last = i + 1 == len(wells)
        if is_last:
            await robot_api.upper_module.pipette.blow_out()
        await robot_api.upper_module.pipette.move_to(
            location=location(
                location_type=LocationType.CHIP_HOLDER,
                location_number=1,
                well=well,
                reference=LocationReference.TOP_JUST,
            ),
            z_speed=Speed.from_mm(10),
            optimize=True,
        )

    # deck_module.inspector
    await robot_api.deck_module.inspector.set_lcr_param(
        value=LCRParam(
            data_format_long=False,
            frequency=5000,
            voltage_level=2,
            measurement_time_mode=MeasurementTimeMode.MEDIUM,
            averaging_rate=1,
        )
    )
    lcr_values: dict[int, float] = await robot_api.deck_module.inspector.scan()
    capacitance_values, impedance_values = (
        await robot_api.deck_module.inspector.scan_capacitance_impedance(
            capacitance_repeat_count=0,
            impedance_repeat_count=1,
        )
    )
    await robot_api.deck_module.inspector.prepare_scan()
    # for i in range(192):
    for i in range(3):
        await robot_api.deck_module.inspector.set_switch_on(channel=i + 1)
        lcr_value: float = await robot_api.deck_module.inspector.get_lcr_value()
    await robot_api.deck_module.inspector.complete_scan()

    # deck_module.heater
    heater_on: dict[int, bool] = await robot_api.deck_module.heater.get_heater_on()
    heater_temperature: dict[int, float] = (
        await robot_api.deck_module.heater.get_temperature()
    )
    await robot_api.deck_module.heater.set_temperature(
        values={
            1: HeaterData(on=True, temperature=37.0),
            2: HeaterData(on=False),
            # ...
            8: HeaterData(on=True, temperature=37.5),
        }
    )

    # axis
    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)


if __name__ == "__main__":
    # asyncio.run(main())
    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



from ablelabs.neon_v2.notable.core import Base
from ablelabs.neon_v2.notable.enums import LEDColor


class IO(Base):
    """Input/Output control for robot sensors, indicators, and digital interfaces.
    
    This class manages all I/O operations including digital inputs/outputs,
    environmental sensors, LED controls, and door monitoring. It provides
    real-time monitoring and control of the robot's peripheral systems.
    """
    # Input operations
    async def get_inputs(self) -> dict:
        """Get all digital input (DI) port states.
        
        Returns:
            dict: Current state of all digital input ports
                Format: {"port_1": bool, "port_2": bool, ...}
        """
        return await self._get("/api/v1/driver/io/inputs")

    async def get_outputs(self) -> dict:
        """Get all digital output (DO) port states."""
        return await self._get("/api/v1/driver/io/outputs")

    async def set_outputs(self, outputs: dict) -> dict:
        """Set multiple digital output (DO) pin states."""
        return await self._post("/api/v1/driver/io/outputs", body=outputs)

    async def get_pdo(self, pin) -> float:
        """Get PDO pin PWM state (0-100)."""
        return await self._get("/api/v1/driver/io/pdo", params={"pin": pin})

    async def set_pdo(self, pin: int, value: float | bool) -> dict:
        """Set PDO pin state (ON/OFF or PWM 0-100).
        
        Args:
            pin: PDO pin number
            value: Pin state - bool for ON/OFF, or float (0-100) for PWM
            
        Returns:
            dict: PDO setting confirmation
        """
        return await self._post(
            "/api/v1/driver/io/pdo", params={"pin": pin, "value": value}
        )

    async def get_environment(self) -> dict:
        """Get environment sensor data (temperature, humidity, pressure).
        
        Returns:
            dict: Environmental measurements containing:
                - temperature: Temperature in Celsius
                - humidity: Relative humidity percentage
                - pressure: Atmospheric pressure in hPa
        """
        return await self._get("/api/v1/driver/io/environment")

    async def get_door(self) -> bool:
        """Get door open/closed status.
        
        Returns:
            bool: True if door is open, False if closed
            
        Note:
            Door should be closed during robot operation for safety
        """
        return await self._get("/api/v1/driver/io/door")

    # LED controls
    async def set_led_lamp(self, on: bool) -> None:
        """Turn LED lamp on or off."""
        return await self._post("/api/v1/driver/io/led-lamp", params={"on": on})

    async def set_led_bar(
        self,
        color: LEDColor = LEDColor.GREEN,
        bright_percent: int = 20,
        progress_percent: int = 100,
        blink_time_ms: int = 0,
    ) -> None:
        """Set LED bar color, brightness, progress, and blinking.
        
        Args:
            color: LED color from LEDColor enum
            bright_percent: Brightness level (0-100)
            progress_percent: Progress bar fill percentage (0-100)
            blink_time_ms: Blink interval in milliseconds (0 = no blinking)
            
        Example:
            # Set blue LED at 50% brightness, full progress, no blinking
            await io.set_led_bar(color="BLUE", bright_percent=50, progress_percent=100, blink_time_ms=0)
        """
        params = {
            "color": color,
            "bright_percent": bright_percent,
            "progress_percent": progress_percent,
            "blink_time_ms": blink_time_ms,
        }
        return await self._post("/api/v1/driver/io/led-bar", params=params)

    async def set_led_bar_r(self, percent: int) -> None:
        """Set LED bar red brightness (0-100)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/r", params={"percent": percent}
        )

    async def set_led_bar_g(self, percent: int) -> None:
        """Set LED bar green brightness (0-100)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/g", params={"percent": percent}
        )

    async def set_led_bar_b(self, percent: int) -> None:
        """Set LED bar blue brightness (0-100)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/b", params={"percent": percent}
        )

    async def set_led_bar_w(self, percent: int) -> None:
        """Set LED bar white brightness (0-100)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/w", params={"percent": percent}
        )

    async def set_led_bar_percent(self, percent: int) -> None:
        """Set LED bar progress display (0-100)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/percent", params={"percent": percent}
        )

    async def set_led_bar_blink(self, msec: int) -> None:
        """Set LED bar blink time in milliseconds (0 = no blink)."""
        return await self._post(
            "/api/v1/driver/io/led-bar/blink", params={"msec": msec}
        )



from ablelabs.neon_v2.notable.core import Base


class DriverParam(Base):
    """Driver parameter management for low-level robot configuration.
    
    This class manages driver-level parameters that control the fundamental
    behavior of robot axes, motors, and sensors. These parameters include
    motion profiles, sensor thresholds, and hardware-specific settings.
    """

    async def get_driver_param(self) -> dict:
        """Get all driver parameters.
        
        Retrieves the complete set of driver parameters used by the robot's
        motion control system, sensors, and hardware interfaces.
        
        Returns:
            dict: Complete driver parameter configuration containing:
                - axis_settings: Motion parameters for each axis
                - sensor_config: Sensor calibration and threshold values  
                - hardware_limits: Safety limits and constraints
                - timing_parameters: Motion timing and synchronization settings
        """
        return await self._get("/api/v1/resource/driver-param/")

    async def get_driver_param_keys(self, keys: list[str]) -> dict:
        """Get specific driver parameter values by keys.
        
        Retrieves only the requested parameter values, which is more efficient
        than loading the entire parameter set when only specific values are needed.
        
        Args:
            keys: List of parameter keys to retrieve
            
        Returns:
            dict: Requested parameter values keyed by parameter name
            
        Example:
            keys = ["axis.x.max_speed", "axis.z1.acceleration", "sensor.door.threshold"]
            params = await driver_param.get_driver_param_keys(keys)
        """
        return await self._post(
            "/api/v1/resource/driver-param/keys", body={"keys": keys}
        )

    async def reload(self) -> dict:
        """Reload driver parameters from file.
        
        Forces a reload of driver parameters from the configuration file.
        This is useful after manually editing parameter files or during
        system maintenance and calibration procedures.
        
        Returns:
            dict: Reload status and any validation errors
            
        Warning:
            Reloading parameters may affect robot motion behavior.
            Only perform this operation when the robot is idle and safe.
        """
        return await self._post("/api/v1/resource/driver-param/reload")

    async def update(self, data: dict) -> dict:
        """Update driver parameter values.
        
        Modifies specific driver parameters with new values. Changes take
        effect immediately and persist until the next system restart.
        
        Args:
            data: Dictionary of parameter keys and new values to update
            
        Returns:
            dict: Update confirmation and any validation warnings
            
        Example:
            data = {
                "axis.x.max_speed": 150.0,      # mm/s
                "axis.z1.acceleration": 500.0,   # mm/s²
                "pipette.1.flow_rate_max": 400.0 # μL/s
            }
            
        Warning:
            Incorrect parameter values can cause unsafe robot behavior.
            Only modify parameters if you understand their effects.
        """
        return await self._post("/api/v1/resource/driver-param/update", body=data)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.common.alma.structs import LCRParam


class InspectorAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.deck_module_inspector_set_lcr_param)
    async def set_lcr_param(self, value: LCRParam):
        pass

    @run_server_func(RobotRouter.deck_module_inspector_scan)
    async def scan(self) -> dict[int, float]:
        pass

    @run_server_func(RobotRouter.deck_module_inspector_scan_capacitance_impedance)
    async def scan_capacitance_impedance(
        self,
        capacitance_repeat_count: int,
        impedance_repeat_count: int,
    ) -> tuple[dict[int, list[float]], dict[int, list[float]]]:
        pass

    @run_server_func(RobotRouter.deck_module_inspector_prepare_scan)
    async def prepare_scan(self):
        pass

    @run_server_func(RobotRouter.deck_module_inspector_complete_scan)
    async def complete_scan(self):
        pass

    @run_server_func(RobotRouter.deck_module_inspector_set_switch_on)
    async def set_switch_on(self, channel: int):
        pass

    @run_server_func(RobotRouter.deck_module_inspector_set_switch_off)
    async def set_switch_off(self):
        pass

    @run_server_func(RobotRouter.deck_module_inspector_get_lcr_value)
    async def get_lcr_value(self) -> float:
        pass



from dataclasses import dataclass

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.structs import Speed, FlowRate, LedBarParam
from ablelabs.neon.common.notable.nanophilia.enums import (
    LocationType,
    LocationReference,
)


@dataclass
class Location:
    location_type: LocationType
    location_number: int = None
    well: str = None
    reference: LocationReference = None
    offset: tuple[float, float, float] = (0, 0, 0)

    def __str__(self):
        # 필드값이 기본값과 다를 때만 표시
        fields = []
        if self.location_number == None:
            fields.append(f"{self.location_type}")
        else:
            fields.append(f"{self.location_type}.{self.location_number}")
        if self.well != None:
            fields.append(f"well={self.well}")
        if self.reference != None:
            fields.append(f"reference={self.reference}")
        if self.offset != (0, 0, 0):
            fields.append(f"offset={self.offset}")
        return f"Location({', '.join(fields)})"


def location(
    location_type: LocationType = LocationType.DECK,
    location_number: int = None,
    well: str = None,
    reference: LocationReference = None,
    offset: tuple[float] = (0, 0, 0),
):
    return Location(
        location_type=location_type,
        location_number=location_number,
        well=well,
        reference=reference,
        offset=offset,
    )



from loguru import logger
from typing import Literal

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.api.set_api import SetAPI
from ablelabs.neon.controllers.notable.api.get_api import GetAPI
from ablelabs.neon.controllers.notable.api.motion_api import MotionAPI
from ablelabs.neon.controllers.notable.api.axis_api import AxisAPI
from ablelabs.neon.common.notable.enums import RunStatus


class RobotAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient = None) -> None:
        if tcp_client == None:
            tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)
        super().__init__(tcp_client)
        self._set_api = SetAPI(tcp_client=tcp_client)
        self._get_api = GetAPI(tcp_client=tcp_client)
        self._motion_api = MotionAPI(tcp_client=tcp_client)
        self._axis_api = AxisAPI(tcp_client=tcp_client)

    @property
    def set(self):
        return self._set_api

    @property
    def get(self):
        return self._get_api

    @property
    def motion(self):
        return self._motion_api

    @property
    def axis(self):
        return self._axis_api

    async def connect(self, ip, port):
        await self._tcp_client.connect(ip=ip, port=port)

    @run_server_func(RobotRouter.robot_wait_boot)
    async def wait_boot(self):
        pass

    @run_server_func(RobotRouter.robot_stop)
    async def stop(self):
        pass

    @run_server_func(RobotRouter.robot_clear_error)
    async def clear_error(self):
        pass

    @run_server_func(RobotRouter.robot_pause)
    async def pause(self):
        pass

    @run_server_func(RobotRouter.robot_resume)
    async def resume(self):
        pass

    @run_server_func(RobotRouter.robot_is_connected)
    async def is_connected(self) -> dict[Literal["dio", "motor"], bool]:
        pass

    @run_server_func(RobotRouter.robot_get_run_status)
    async def get_run_status(self) -> RunStatus:
        pass

    @run_server_func(RobotRouter.robot_set_run_status)
    async def set_run_status(self, value: RunStatus):
        pass



from ablelabs.neon_v2.notable.core import Base


class Library(Base):
    """Resource library management for pipettes and labware.
    
    This class manages the robot's knowledge base of available pipettes and labware,
    including their physical dimensions, properties, and capabilities. Libraries are
    loaded from JSON files and can be updated or reloaded as needed.
    """

    async def get_pipette(self) -> dict:
        """Get pipette library information.
        
        Retrieves complete pipette library with specifications for all
        available pipette types including volume ranges, flow rates,
        and physical dimensions.
        
        Returns:
            dict: Pipette library mapping pipette codes to specifications.
                Format: {"pipette_code": {pipette_specs}, ...}
                
        Example:
            {
                "8ch_200ul": {
                    "name": "8-Channel 200μL",
                    "channels": 8,
                    "min_volume": 20,
                    "max_volume": 200,
                    "flow_rate": {"min": 1, "max": 500}
                }
            }
        """
        return await self._get("/api/v1/resource/library/pipette")

    async def update(self, pipette_code: str, data: dict) -> dict:
        """Update specific pipette library data.
        
        Modifies the specifications for a specific pipette type in the library.
        This allows customization of pipette parameters for specific applications.
        
        Args:
            pipette_code: Code identifying the pipette type to update
            data: Dictionary containing the updated pipette specifications
            
        Returns:
            dict: Confirmation of the library update
            
        Example:
            data = {
                "flow_rate": {"min": 5, "max": 300},  # Updated flow rate range
                "custom_offset": [0, 0, 2.5]          # Custom Z offset
            }
        """
        return await self._post(
            "/api/v1/resource/library/pipette/update",
            params={"pipette_code": pipette_code},
            body=data,
        )

    async def get_labware(self) -> dict:
        """Get labware library information.
        
        Retrieves complete labware library with specifications for all
        available labware types including dimensions, well layouts,
        and volume capacities.
        
        Returns:
            dict: Labware library organized by categories.
                Format: {"category": {"labware_code": {specs}, ...}, ...}
                
        Example:
            {
                "tip_box": {
                    "ablelabs_tip_box_200": {
                        "name": "ABLE Labs 200μL Tip Box",
                        "wells": 96,
                        "well_volume": 200
                    }
                },
                "well_plate": {
                    "spl_96_well_plate_30096": {
                        "name": "96-Well Plate",
                        "wells": 96,
                        "well_volume": 200
                    }
                }
            }
        """
        return await self._get("/api/v1/resource/library/labware")

    async def reload(self) -> dict:
        """Reload pipette and labware libraries from files.
        
        Forces a reload of both pipette and labware libraries from their
        respective JSON files. This is useful after manually editing library
        files or after software updates.
        
        Returns:
            dict: Reload status and any errors encountered
            
        Note:
            This operation may take several seconds to complete as it
            re-parses all library files and validates the data
        """
        return await self._post("/api/v1/resource/library/reload")



import asyncio

# Import from new structure
from ablelabs.neon_v2.notable import Notable


async def main():
    import subprocess

    # Launch Neon.exe using subprocess (for Windows systems)
    # Ensure 'Neon.exe' and the 'resources' folder are in the correct directory
    subprocess.Popen(
        [
            r"C:\your_path\Neon.exe",
        ]
    )

    base_url = "http://localhost:7777"
    notable = Notable(base_url)

    # configure pippettes
    await notable.config.pipette.set_pipette({"1": "8ch_200ul"})

    # configure deck
    await notable.config.deck.set_deck(
        {
            "7": "ablelabs_tip_box_200",
            "8": "nest_12_reservoir_360102",
            "9": "spl_96_well_plate_30096",
            "12": "ablelabs_trash",
        }
    )
    # initialize robot
    await notable.controller.upper_module.initialize(home_axes=True, move_to_ready=True)

    # pick up tip, drop tip example
    status = await notable.action.robot.get_status()
    print(f"Robot Status: {status}")

    # Lamp control and LED bar control
    await notable.driver.io.set_led_lamp(True)
    await notable.driver.io.set_led_bar(
        color="RED", bright_percent=50, progress_percent=50, blink_time_ms=1000
    )

    # Pick up tip
    await notable.controller.upper_module.pick_up_tip(
        pipette_number=1, deck_number=7, well="A1"
    )

    # move to reservoir
    await notable.controller.upper_module.move_to(
        pipette_number=1,
        deck_number=8,
        well="A1",
        z_reference="TOP",
        xyz_offset=[0.0, 0.0, 0.0],
        z_speed=False,
    )

    # aspirate
    await notable.controller.upper_module.aspirate(
        pipette_number=1,
        volume=100.0,
        flow_rate=50.0,
        z_offset=1.0,
        pause_sec=0.0,
    )

    # move to plate
    await notable.controller.upper_module.move_to(
        pipette_number=1,
        deck_number=9,
        well="A1",
        z_reference="TOP",
        xyz_offset=[0.0, 0.0, 0.0],
        z_speed=False,
    )

    await notable.controller.upper_module.dispense(
        pipette_number=1,
        volume=100.0,
        flow_rate=50.0,
        z_offset=1.0,
        pause_sec=0.0,
    )

    # Drop tip
    await notable.controller.upper_module.drop_tip(
        pipette_number=1, deck_number=7, well="A1"
    )  # tip_box
    await notable.controller.upper_module.drop_tip(
        pipette_number=1, deck_number=12
    )  # trash


if __name__ == "__main__":
    asyncio.run(main())



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter


class GetAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.get_pipette_infos)
    async def pipette_infos(self) -> dict[int, dict]:
        pass

    @run_server_func(RobotRouter.get_tip_infos)
    async def tip_infos(self) -> dict[int, dict]:
        pass

    @run_server_func(RobotRouter.get_deck_module_infos)
    async def deck_module_infos(self) -> dict[int, dict]:
        pass

    @run_server_func(RobotRouter.get_labware_infos)
    async def labware_infos(self) -> dict[int, dict]:
        pass

    @run_server_func(RobotRouter.get_setup_data)
    async def setup_data(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_region_dataset)
    async def region_dataset(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_pipette_dataset)
    async def pipette_dataset(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_tip_dataset)
    async def tip_dataset(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_deck_module_dataset)
    async def deck_module_dataset(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_labware_dataset)
    async def labware_dataset(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_region_json)
    async def region_json(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_pipette_json)
    async def pipette_json(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_tip_json)
    async def tip_json(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_deck_module_json)
    async def deck_module_json(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_labware_json)
    async def labware_json(self) -> dict:
        pass

    @run_server_func(RobotRouter.get_setup_data_toml)
    async def setup_data_toml(self) -> str:
        pass

    @run_server_func(RobotRouter.get_driver_param_toml)
    async def driver_param_toml(self) -> str:
        pass



from ablelabs.neon_v2.notable.core import Base
from ablelabs.neon_v2.notable.enums import Axis as AxisEnum


class Axis(Base):
    """Low-level axis control and motion driver interface.
    
    This class provides direct control over individual robot axes including
    X, Y, Z1, Z2, P1, P2. It offers precise motion control, status monitoring,
    and configuration management for each axis independently.
    
    Available Axes:
        - X: X-axis (horizontal movement)
        - Y: Y-axis (horizontal movement) 
        - Z1: Z-axis for pipette mount 1
        - Z2: Z-axis for pipette mount 2
        - P1: Plunger axis for pipette 1
        - P2: Plunger axis for pipette 2
    """
    async def get_max_speed(self) -> dict:
        """Get maximum speed limits for all axes.
        
        Returns:
            dict: Maximum speed values for each axis in mm/s or steps/s
        """
        return await self._get("/api/v1/driver/axis/max-speed")

    async def get_status(self, axis: AxisEnum) -> dict[str, bool]:
        """Get current status of specific axis."""
        return await self._get("/api/v1/driver/axis/status", params={"axis": axis})

    async def get_fault(self, axis: AxisEnum) -> dict[str, bool]:
        """Get fault status for specific axis."""
        return await self._get("/api/v1/driver/axis/fault", params={"axis": axis})

    async def get_error_code(self, axis: AxisEnum) -> str:
        """Get error code for specific axis."""
        return await self._get("/api/v1/driver/axis/error-code", params={"axis": axis})

    async def is_enabled(self, axis: AxisEnum) -> bool:
        """Check if specific axis servo is enabled."""
        return await self._get("/api/v1/driver/axis/is-enabled", params={"axis": axis})

    async def is_home_done(self, axis: AxisEnum) -> bool:
        """Check if homing is completed for specific axis."""
        return await self._get(
            "/api/v1/driver/axis/is-home-done", params={"axis": axis}
        )

    async def is_moving(self, axis: AxisEnum) -> bool:
        """Check if specific axis is currently moving."""
        return await self._get("/api/v1/driver/axis/is-moving", params={"axis": axis})

    async def is_move_done(self, axis: AxisEnum) -> bool:
        """Check if last movement command is completed for specific axis."""
        return await self._get(
            "/api/v1/driver/axis/is-move-done", params={"axis": axis}
        )

    async def get_position(self, axis: AxisEnum, unit: bool = True) -> float:
        """Get current position of specific axis.
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            unit: If True, return position in mm; if False, return in steps
            
        Returns:
            dict: Current axis position information
        """
        return await self._get(
            "/api/v1/driver/axis/position", params={"axis": axis, "unit": unit}
        )

    async def set_position(self, axis: AxisEnum, position: float, unit: bool = True) -> None:
        """Force set current position value without actual movement."""
        params = {"axis": axis, "position": position, "unit": unit}
        return await self._post("/api/v1/driver/axis/position", params=params)

    async def get_home_offset(self, axis: AxisEnum) -> float:
        """Get home offset value for specific axis."""
        return await self._get("/api/v1/driver/axis/home-offset", params={"axis": axis})

    async def set_home_offset(self, axis: AxisEnum, home_offset: float) -> None:
        """Set home offset value for specific axis."""
        params = {"axis": axis, "home_offset": home_offset}
        return await self._post("/api/v1/driver/axis/home-offset", params=params)

    async def get_resolution(self, axis: AxisEnum) -> int:
        """Get resolution value for specific axis."""
        return await self._get("/api/v1/driver/axis/resolution", params={"axis": axis})

    async def set_resolution(self, axis: AxisEnum, resolution: float) -> None:
        """Set resolution value for specific axis."""
        params = {"axis": axis, "resolution": resolution}
        return await self._post("/api/v1/driver/axis/resolution", params=params)

    async def enable(self, axis: AxisEnum) -> None:
        """Enable axis servo."""
        return await self._post("/api/v1/driver/axis/enable", params={"axis": axis})

    async def disable(self, axis: AxisEnum) -> None:
        """Disable axis servo."""
        return await self._post("/api/v1/driver/axis/disable", params={"axis": axis})

    async def clear_fault(self, axis: AxisEnum) -> None:
        """Clear fault state for specific axis."""
        return await self._post(
            "/api/v1/driver/axis/clear-fault", params={"axis": axis}
        )

    async def set_digital_output(self, axis: AxisEnum, channel: int, on: bool) -> None:
        """Control digital output connected to specific axis driver."""
        params = {"axis": axis, "channel": channel, "on": on}
        return await self._post("/api/v1/driver/axis/digital-output", params=params)

    async def stop(self, axis: AxisEnum) -> None:
        """Stop axis movement immediately."""
        return await self._post("/api/v1/driver/axis/stop", params={"axis": axis})

    async def home(self, axis: AxisEnum) -> None:
        """Execute homing operation for specified axis.
        
        Performs precise homing by moving the axis until it reaches
        the home sensor, then sets the home position as reference.
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            
        Note:
            This function executes asynchronously and returns immediately.
            Use wait_home_done() to wait for completion.
        """
        return await self._post("/api/v1/driver/axis/home", params={"axis": axis})

    async def set_speed(self, axis: AxisEnum, speed: float, unit: bool = True) -> None:
        """Set movement speed for specific axis.
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            speed: Movement speed value (must be > 0)
            unit: If True, speed in mm/s; if False, speed in steps/s
            
        Note:
            Speed changes take effect immediately for subsequent movements.
        """
        params = {"axis": axis, "speed": speed, "unit": unit}
        return await self._post("/api/v1/driver/axis/speed", params=params)

    async def set_accel(self, axis: AxisEnum, accel: float, unit: bool = True) -> None:
        """Set acceleration for specific axis."""
        params = {"axis": axis, "accel": accel, "unit": unit}
        return await self._post("/api/v1/driver/axis/accel", params=params)

    async def set_decel(self, axis: AxisEnum, decel: float, unit: bool = True) -> None:
        """Set deceleration for specific axis."""
        params = {"axis": axis, "decel": decel, "unit": unit}
        return await self._post("/api/v1/driver/axis/decel", params=params)

    async def jog(self, axis: AxisEnum, value: float, unit: bool = True) -> None:
        """Move axis continuously at specified speed (0 = stop).
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            value: Jog speed (0 = stop)
                - Positive values: move in positive direction
                - Negative values: move in negative direction
                - Zero: stop movement
            unit: Speed unit selection
                - True: speed in mm/s (or unit/s)
                - False: speed in steps/s
                
        Note:
            Jog continues until stopped by calling jog(axis, 0) or another motion command.
            This function executes asynchronously and returns immediately.
        """
        params = {"axis": axis, "value": value, "unit": unit}
        return await self._post("/api/v1/driver/axis/jog", params=params)

    async def step(self, axis: AxisEnum, value: float, unit: bool = True) -> None:
        """Move axis by relative distance from current position.
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            value: Relative movement distance (positive or negative)
            unit: If True, distance in mm; if False, distance in steps
            
        Returns:
        Note:
            This function executes asynchronously and returns immediately.
            Use wait_move_done() to wait for completion.
        """
        return await self._post(
            "/api/v1/driver/axis/step",
            params={"axis": axis, "value": value, "unit": unit},
        )

    async def move(self, axis: AxisEnum, position: float, unit: bool = True) -> None:
        """Move axis to absolute position.
        
        Args:
            axis: Target axis (X, Y, Z1, Z2, P1, P2)
            position: Target position value
            unit: If True, position in mm; if False, position in steps
            
        Returns:
        Note:
            This function executes asynchronously and returns immediately.
            Use wait_move_done() to wait for completion.
        """
        return await self._post(
            "/api/v1/driver/axis/move",
            params={"axis": axis, "position": position, "unit": unit},
        )

    async def wait_home_done(self, axis: AxisEnum, timeout: float = None) -> None:
        """Wait for homing operation to complete."""
        params = {"axis": axis}
        if timeout is not None:
            params["timeout"] = timeout
        return await self._post("/api/v1/driver/axis/wait-home-done", params=params)

    async def wait_move_done(self, axis: AxisEnum, timeout: float = None) -> None:
        """Wait for axis movement to complete."""
        params = {"axis": axis}
        if timeout is not None:
            params["timeout"] = timeout
        return await self._post("/api/v1/driver/axis/wait-move-done", params=params)

    async def repeat(
        self,
        axis: AxisEnum,
        pos1: float,
        pos2: float,
        delay_ms: float = 0,
        count: int = 1,
        unit: bool = True,
    ) -> None:
        """Move axis back and forth between two positions."""
        params = {
            "axis": axis,
            "pos1": pos1,
            "pos2": pos2,
            "delay_ms": delay_ms,
            "count": count,
            "unit": unit,
        }
        return await self._post("/api/v1/driver/axis/repeat", params=params)

    async def repeats(self, repeats_request: dict, unit: bool = True) -> None:
        """Request multiple axes to repeat movement simultaneously."""
        return await self._post(
            "/api/v1/driver/axis/repeats", params={"unit": unit}, body=repeats_request
        )



from dataclasses import dataclass
import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.structs import Speed, FlowRate, LedBarParam
from ablelabs.neon.common.suitable.enums import LocationType, LocationReference


@dataclass
class Location:
    location_type: LocationType
    location_number: list[int]
    well: list[str]
    reference: LocationReference
    offset: tuple[float] = (0, 0, 0)

    def __post_init__(self):
        if self.location_number != None and self.well != None:
            assert len(self.location_number) == len(
                self.well
            ), f"not same len : location_number={self.location_number} well={self.well}"
        elif self.location_number == None and self.well != None:
            self.location_number = [None] * len(self.well)


def location(
    location_type: LocationType,
    location_number: list[int] = None,
    well: list[str] = None,
    reference: LocationReference = None,
    offset: tuple[float] = (0, 0, 0),
):
    return Location(
        location_type=location_type,
        location_number=location_number,
        well=well,
        reference=reference,
        offset=offset,
    )


if __name__ == "__main__":
    loc = location(
        location_type=LocationType.DECK,
        location_number=[4, 4, 4],
        well=["a1", "b1", "c1"],
    )
    loc = location(
        location_type=LocationType.DECK,
        location_number=[22, 22, 16, 16, 10, 10, 4, 4],
        well=["a1", "c1", "a1", "c1", "a1", "c1", "a1", "c1"],
        reference=LocationReference.TOP,
    )



from ablelabs.neon_v2.notable.core import Base


class Deck(Base):
    """Deck configuration management for labware positioning.
    
    This class manages the robot deck layout configuration, allowing users to
    define which labware types are placed at each deck position (1-12).
    """

    async def get_deck(self) -> dict:
        """Get current deck configuration.
        
        Retrieves the current labware configuration for all deck positions.
        Each position (1-12) can contain a specific labware type or be empty.
        
        Returns:
            dict: Current deck configuration mapping deck positions to labware types.
                Format: {"1": "labware_type", "2": "another_labware", ...}
                Empty positions are not included in the response.
                
        Example:
            {
                "1": "ablelabs_tip_box_200",
                "2": "nest_12_reservoir_360102", 
                "3": "spl_96_well_plate_30096",
                "12": "ablelabs_trash"
            }
        """
        return await self._get("/api/v1/config/deck/")

    async def set_deck(self, config: dict[str | int, str]) -> dict:
        """Set deck configuration for labware positioning.
        
        Configures which labware codes are placed at specific deck positions.
        This configuration is used by the robot to calculate correct positioning
        and movement paths for liquid handling operations.
        
        Args:
            config: Dictionary mapping deck positions to labware codes.
                   Positions can be strings ("1"-"12") or integers (1-12).
                   Values must be valid labware codes from the labware library.
                   
        Returns:
            dict: Complete deck configuration with all positions 1-12.
                 Format: {"1": "labware_code", "2": None, ..., "12": "trash_code"}
                 Unassigned positions return None.
            
        Example:
            config = {
                1: "ablelabs_tip_box_200",        # Tip box at position 1 (int key)
                "2": "nest_12_reservoir_360102",  # Reservoir at position 2 (str key)  
                3: "spl_96_well_plate_30096",     # 96-well plate at position 3
                12: "ablelabs_trash"              # Trash at position 12
            }
            
            # Returns all 12 positions:
            {
                "1": "ablelabs_tip_box_200",
                "2": "nest_12_reservoir_360102", 
                "3": "spl_96_well_plate_30096",
                "4": None,
                ...
                "12": "ablelabs_trash"
            }
            
        Note:
            - Position keys can be strings or integers (both accepted)
            - Positions must be in range 1-12
            - Labware codes must exist in the labware library
            - Invalid labware codes will cause API errors
            - Empty positions should be omitted from input (will return None)
        """
        return await self._post("/api/v1/config/deck/", body=config)


from typing import Any

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerServer
from ablelabs.neon.common.suitable.enums import Axis, RunStatus, PipetteCalibrationType
from ablelabs.neon.common.suitable.structs import Location, Speed, FlowRate


class RobotRouter(MessengerServer):
    async def robot_wait_boot(self):
        pass

    async def robot_stop(self):
        pass

    async def robot_clear_error(self):
        pass

    async def robot_pause(self):
        pass

    async def robot_resume(self):
        pass

    async def robot_is_connected(self):
        pass

    async def robot_get_environment(self):
        pass

    async def robot_get_run_status(self):
        pass

    async def robot_get_changed_robot_status(self):
        pass

    async def robot_set_run_status(self, value: RunStatus):
        pass

    async def robot_set_progress_rate(self, value: float):
        pass

    async def robot_set_interlock(self, value: bool):
        pass

    # get api
    async def get_setup_data(self) -> dict:
        pass

    async def get_position(
        self, labware_code: str = None, deck: int = None, tip_volume: int = None
    ) -> dict:
        pass

    # set api
    async def set_pipettes(self, value: dict[int, str | None]):
        pass

    async def set_tips(self, value: dict[int, str | None]):
        pass

    async def set_labwares(self, value: dict[int, str | None]):
        pass

    async def set_update_setup_data(self, value: dict, save: bool):
        pass

    async def set_update_pipette_attrs(self, value: dict[int, dict]):
        pass

    async def set_pipette_calibrations(self, value: dict[int, PipetteCalibrationType]):
        pass

    async def set_position(
        self,
        labware_code: str,
        deck: int,
        x: float,
        ys: list[float],
        zs: list[float],
        tip_volume: int = None,
        save: bool = True,
    ) -> dict:
        pass

    # state api
    async def state_get_current_motion(self):
        pass

    async def state_get_estimated_time(self):
        pass

    # time api
    async def time_initialize(self):
        pass

    async def time_delay(self, sec: float):
        pass

    async def time_move_to_ready(self):
        pass

    async def time_move_to(
        self,
        pipette_number: list[int],
        location: Location,
    ):
        pass

    async def time_pick_up_tip(
        self,
        pipette_number: list[int],
        location: Location,
    ):
        pass

    async def time_drop_tip(
        self,
        pipette_number: list[int],
        location: Location,
    ):
        pass

    async def time_rise_tip(
        self,
        pipette_number: list[int],
        height_offset: float | list[float],
        z_speed: Speed | list[Speed],
    ):
        pass

    async def time_aspirate(
        self,
        pipette_number: list[int],
        volume: (
            float | list[list[float]]
        ),  # [[p1.v1, p2.v1], [p1.v2, p2.v2], [p1.v3, p2.v3], ... ]
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        rise_tip_height_offset: float = None,
        rise_tip_speed: Speed = Speed.from_mm(10),
        pre_wet_count: int = 0,
    ):
        pass

    async def time_dispense(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        rise_tip_height_offset: float = None,
        rise_tip_speed: Speed = Speed.from_mm(10),
    ):
        pass

    async def time_mix(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        delay: float = 0.0,
        rise_tip_height_offset: float = None,
        rise_tip_speed: Speed = Speed.from_mm(10),
    ):
        pass

    async def time_blow_out(
        self,
        pipette_number: list[int],
        flow_rate: FlowRate | list[FlowRate] = None,
    ):
        pass

    # motion api
    async def motion_initialize(self):
        pass

    async def motion_home_x(self):
        pass

    async def motion_home_y(self):
        pass

    async def motion_home_z(self):
        pass

    async def motion_home_p(self):
        pass

    async def motion_delay(self, sec: float):
        pass

    async def motion_move_z_up(self):
        pass

    async def motion_move_to_ready(self):
        pass

    async def motion_move_to(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_pick_up_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_drop_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_rise_tip(
        self,
        pipette_number: list[int],
        height_offset: float | list[float],
        z_speed: Speed | list[Speed],
        optimize: bool = False,
    ):
        pass

    async def motion_aspirate(
        self,
        pipette_number: list[int],
        volume: float | list[list[float]],
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        rise_tip_height_offset: float = None,
        pre_wet_count: int = 0,
        pre_wet_volume: float | list[float] = None,
        optimize: bool = False,
    ):
        pass

    async def motion_dispense(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        liquid_following: bool = False,
        rise_tip_height_offset: float = None,
        optimize: bool = False,
    ):
        pass

    async def motion_mix(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate | list[FlowRate] = None,
        delay: float = 0.0,
        liquid_following: bool = False,
        rise_tip_height_offset: float = None,
        optimize: bool = False,
    ):
        pass

    async def motion_blow_out(
        self,
        pipette_number: list[int],
        flow_rate: FlowRate | list[FlowRate] = None,
    ):
        pass

    # axis api
    async def axis_get_position(self, axis: Axis, floor_digit: int = 1):
        pass

    async def axis_set_speed(self, axis: Axis, value: float):
        pass

    async def axis_set_accel(self, axis: Axis, value: float):
        pass

    async def axis_set_decel(self, axis: Axis, value: float):
        pass

    async def axis_enable(self, axis: Axis):
        pass

    async def axis_disable(self, axis: Axis):
        pass

    async def axis_stop(self, axis: Axis):
        pass

    async def axis_home(self, axis: Axis):
        pass

    async def axis_jog(self, axis: Axis, value: float):
        pass

    async def axis_step(self, axis: Axis, value: float):
        pass

    async def axis_move(self, axis: Axis, value: float):
        pass

    async def axis_wait_home_done(self, axis: Axis):
        pass

    async def axis_wait_move_done(self, axis: Axis):
        pass



from loguru import logger

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.kribb.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.api.set_api import SetAPI as BaseSetAPI


class SetAPI(BaseSetAPI):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.set_camera)
    async def camera(
        self,
        pipette_number: int,
    ):
        pass



from ablelabs.neon_v2.notable.modules import Notable
from ablelabs.neon_v2.notable.enums import (
    LEDColor,
    ZReference,
    Axis,
    PipetteType,
    LabwareType,
)

__all__ = ["Notable", "LEDColor", "ZReference", "Axis", "PipetteType", "LabwareType"]


from ablelabs.neon_v2.notable.core import Base


class SetupData(Base):
    """Setup data management for robot system configuration.
    
    This class manages the robot's system configuration data stored in
    setup_data.toml file, including communication settings, hardware
    configuration, calibration data, and operational parameters.
    """

    async def get_setup_data(self) -> dict:
        """Get all setup data.
        
        Retrieves the complete setup configuration including hardware settings,
        communication parameters, calibration data, and system preferences.
        
        Returns:
            dict: Complete setup data configuration containing:
                - communication: Serial port and network settings
                - env: Environment and operational mode settings
                - hardware: Hardware-specific configuration parameters
                - calibration: Calibration offsets and teaching positions
                - safety: Safety limits and emergency stop settings
                
        Example:
            {
                "communication": {"port": "COM3", "baudrate": 115200},
                "env": {"is_online": true, "debug_mode": false},
                "hardware": {"pipette_count": 2, "axis_count": 6}
            }
        """
        return await self._get("/api/v1/resource/setup-data/")

    async def get_setup_data_keys(self, keys: list[str]) -> dict:
        """Get specific setup data values by keys.
        
        Retrieves only the requested configuration values, which is more
        efficient than loading the entire setup data when only specific
        values are needed.

        Args:
            keys: List of dot-notation keys to retrieve from setup data
                 (e.g., "communication.port", "env.is_online")

        Returns:
            dict: Dictionary containing requested key-value pairs
            
        Example:
            keys = ["communication.port", "env.is_online", "hardware.pipette_count"]
            config = await setup_data.get_setup_data_keys(keys)
            # Returns: {"communication.port": "COM3", "env.is_online": true, ...}
        """
        return await self._post("/api/v1/resource/setup-data/keys", body={"keys": keys})

    async def reload(self) -> dict:
        """Reload setup_data.toml file into memory.
        
        Forces a reload of the setup data configuration from the TOML file.
        This is useful after manually editing the setup_data.toml file or
        during system reconfiguration.
        
        Returns:
            dict: Reload status and any parsing errors
            
        Note:
            Changes to communication settings may require a system restart
            to take full effect. Always verify connectivity after reloading.
        """
        return await self._post("/api/v1/resource/setup-data/reload")

    async def update(self, data: dict) -> dict:
        """Update setup data values.
        
        Modifies specific setup data parameters with new values. Changes
        are applied immediately and persisted to the configuration file.
        
        Args:
            data: Dictionary of configuration keys and new values to update
                 Use dot notation for nested keys (e.g., "communication.port")
        
        Returns:
            dict: Update confirmation and any validation warnings
            
        Example:
            data = {
                "communication.port": "COM4",
                "env.is_online": True,
                "hardware.emergency_stop_enabled": True
            }
            
        Warning:
            Incorrect configuration values can cause system malfunction.
            Always validate settings before applying updates, especially
            for communication and safety-related parameters.
        """
        return await self._post("/api/v1/resource/setup-data/update", body=data)



from functools import wraps

def extract_data_async(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        response = await func(*args, **kwargs)

        if isinstance(response, dict) and "data" in response:
            return response["data"]

        return response

    return wrapper



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.common.alma.structs import HeaterData


class HeaterAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.deck_module_heater_get_heater_on)
    async def get_heater_on(self) -> dict[int, bool]:
        pass

    @run_server_func(RobotRouter.deck_module_heater_get_temperature)
    async def get_temperature(self) -> dict[int, float]:
        pass

    @run_server_func(RobotRouter.deck_module_heater_set_environment_temperature)
    async def set_environment_temperature(self, value: float):
        pass

    @run_server_func(RobotRouter.deck_module_heater_set_temperature)
    async def set_temperature(self, values: dict[int, HeaterData]):
        pass



# Import from new structure
from ablelabs.neon_v2.notable.resource.driver_param import DriverParam
from ablelabs.neon_v2.notable.resource.setup_data import SetupData
from ablelabs.neon_v2.notable.resource.library import Library
from ablelabs.neon_v2.notable.config.pipette import Pipette
from ablelabs.neon_v2.notable.config.deck import Deck
from ablelabs.neon_v2.notable.driver.io import IO
from ablelabs.neon_v2.notable.driver.axis import Axis
from ablelabs.neon_v2.notable.controller.upper_module import UpperModule
from ablelabs.neon_v2.notable.action.robot import Robot


class Notable:
    def __init__(self, base_url: str):
        self._base_url = base_url

        self.resource = Resource(base_url)
        self.config = Config(base_url)
        self.driver = Driver(base_url)
        self.controller = Controller(base_url)
        self.action = Action(base_url)


class Resource:
    def __init__(self, base_url: str):
        self.driver_param = DriverParam(base_url)
        self.setup_data = SetupData(base_url)
        self.library = Library(base_url)


class Config:
    def __init__(self, base_url: str):
        self.pipette = Pipette(base_url)
        self.deck = Deck(base_url)


class Driver:
    def __init__(self, base_url: str):
        self.io = IO(base_url)
        self.axis = Axis(base_url)


class Controller:
    def __init__(self, base_url: str):
        self.upper_module = UpperModule(base_url)


class Action:
    def __init__(self, base_url: str):
        self.robot = Robot(base_url)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.controllers.alma.api.upper_modules.gripper import GripperAPI
from ablelabs.neon.controllers.alma.api.upper_modules.pipette import PipetteAPI


class UpperModuleAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)
        self.pipette = PipetteAPI(tcp_client)
        self.gripper = GripperAPI(tcp_client)

    @run_server_func(RobotRouter.upper_module_move_z_up)
    async def move_z_up(self):
        pass

    @run_server_func(RobotRouter.upper_module_move_to_ready)
    async def move_to_ready(self):
        pass



# python 3.10 버전

# 가상환경 설정
## python -m venv .venv
## py 3.10 -m venv .venv

# ablelabs 패키지 설치
## pip install ablelabs

# Neon.exe + resources 파일 위치 확인
## resources 폴더는 Neon.exe와 같은 폴더 또는 하위 폴더에

# subprocess.Popen으로 Neon.exe 실행

# connect -> wait_boot -> set_pipettes/tips/labwares 순서대로 호출


import asyncio
import sys, os
from typing import Literal

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.controllers.notable.nanophilia.api.robot_api import RobotAPI
from ablelabs.neon.common.notable.nanophilia.enums import (
    RunStatus,
    LocationType,
    Axis,
    LocationReference,
    DioInput,
    DioPDO,
    AxisDO,
)
from ablelabs.neon.common.notable.nanophilia.structs import Speed, FlowRate, location
from ablelabs.neon.utils.format_conversion import floor_decimal
from ablelabs.neon.utils.location_conversion import LocationConversion, NUMBERING_ORDER


async def main():
    import subprocess

    # for windows.
    # subprocess.Popen(
    #     [
    #         r"C:\your_path\Neon.exe",
    #     ]
    # )

    # for mac.
    subprocess.Popen(
        [
            r"/Users/sypark/Code/ABLE-Elba/.venv/bin/python",
            r"/Users/sypark/Code/ABLE-Elba/robot/src/controllers/notable/nanophilia/robot_router.py",
        ],
        cwd=r"/Users/sypark/Code/ABLE-Elba",
    )
    await asyncio.sleep(2)

    ip = "localhost"
    port = 1234

    robot_api = RobotAPI()
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass
    await robot_api.wait_boot()
    await robot_api.shutdown()

    # set
    await robot_api.set.pipettes(
        {
            1: "8ch200ul",
            2: "1ch200ul",
        }
    )
    await robot_api.set.tips(
        {
            1: "tip_200",
            2: "tip_200",
        }
    )
    await robot_api.set.deck_modules(
        {
            10: "magnetic_shaker",
            7: "shaker",
        }
    )
    await robot_api.set.labwares(
        {
            12: "trash_#v2.5",
            5: "megarobo_tipbox_#MCT200-N-TP-S-B",
            6: "megarobo_tipbox_#MCT200-N-TP-S-B",
            8: "megarobo_tipbox_#MCT200-N-TP-S-B",
            9: "megarobo_tipbox_#MCT200-N-TP-S-B",
            1: "spl_96_well_0.2ml_#30096",
            7: "spl_96_well_0.2ml_#30096",
            10: "spl_96_well_0.2ml_#30096",
            2: "NEST_12Channel_15ml_#360102",
        }
    )

    # get
    pipette_infos = await robot_api.get.pipette_infos()
    tip_infos = await robot_api.get.tip_infos()
    deck_module_infos = await robot_api.get.deck_module_infos()
    labware_infos = await robot_api.get.labware_infos()
    setup_data = await robot_api.get.setup_data()

    region_dataset = await robot_api.get.region_dataset()
    pipette_dataset = await robot_api.get.pipette_dataset()
    tip_dataset = await robot_api.get.tip_dataset()
    deck_module_dataset = await robot_api.get.deck_module_dataset()
    labware_dataset = await robot_api.get.labware_dataset()

    region_json = await robot_api.get.region_json()
    pipette_json = await robot_api.get.pipette_json()
    tip_json = await robot_api.get.tip_json()
    deck_module_json = await robot_api.get.deck_module_json()
    labware_json = await robot_api.get.labware_json()
    setup_data_toml = await robot_api.get.setup_data_toml()
    driver_param_toml = await robot_api.get.driver_param_toml()

    x, y, z = setup_data["teaching"]["pipette_1"]["8ch200ul"]["origin_1"]
    setup_data["teaching"]["pipette_1"]["8ch200ul"]["origin_1"] = [x + 1, y + 2, z + 3]
    await robot_api.set.update_setup_data(value=setup_data, save=True)
    await robot_api.set.update_pipette_attrs(
        value={
            1: {
                "blow_out_volume": 20,
            }
        }
    )

    # robot
    await robot_api.stop()
    await robot_api.clear_error()  # stop 이후, motion 전에.
    await robot_api.pause()
    await robot_api.resume()
    is_connected: dict[
        Literal[
            "motor",
            "dio",
            "shaker",
        ],
        bool,
    ] = await robot_api.is_connected()
    dio_input = await robot_api.get_dio_input()
    environment = await robot_api.get_environment()
    run_status = await robot_api.get_run_status()
    changed_robot_status = await robot_api.get_changed_robot_status()
    await robot_api.set_interlock(value=True)
    await robot_api.set_led_lamp(on=True)
    await robot_api.set_run_status(value=RunStatus.INITIALIZE)
    await robot_api.initialize()
    await robot_api.set_run_status(value=RunStatus.RUN)
    await robot_api.set_progress_rate(value=5)
    await robot_api.delay(sec=5)
    await robot_api.set_progress_rate(value=50)
    await robot_api.delay(sec=5)
    await robot_api.set_progress_rate(value=100)
    await robot_api.set_run_status(value=RunStatus.DONE)
    await robot_api.set_run_status(value=RunStatus.READY)

    # FFU
    ffu_status = await robot_api.get_ffu()  # 상태 조회
    await robot_api.set_ffu(on=True, rpm=1000)  # FFU 켜기, RPM 설정
    await robot_api.set_ffu(on=False)  # FFU 끄기
    await robot_api.set_ffu_alarm_clear()  # 알람 클리어

    # cold plate
    coldplate_temperature = await robot_api.get_coldplate_temperature()
    await robot_api.set_coldplate_temperature(on=True, value=-10)
    await robot_api.set_coldplate_temperature(on=False)

    # motion
    await robot_api.motion.initialize()
    await robot_api.motion.initialize_for_washer_dryer()
    await robot_api.motion.move_to(
        pipette_number=1,
        location=location(
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(1.0, 2.0, 3.0),
        ),
    )

    await robot_api.motion.pick_up_tip(
        pipette_number=1,
        location=location(
            location_number=5,
            well="a1",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=1,
        location=location(
            location_number=5,
            well="a1",
        ),
    )
    await robot_api.motion.pick_up_tip(
        pipette_number=1,
        location=location(
            location_number=5,
            well="a12",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=1,
        location=location(
            location_number=12,
        ),
    )

    await robot_api.motion.aspirate(
        pipette_number=1,
        volume=20,
        location=location(
            location_number=2,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.rise_tip(
        pipette_number=1,
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    await robot_api.motion.dispense(
        pipette_number=1,
        volume=20,
        location=location(
            location_number=1,
            well="a2",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.mix(
        pipette_number=1,
        volume=20,
        iteration=2,
        # location=location(
        #     location_number=1,
        #     well="a3",
        #     reference=LocationReference.BOTTOM,
        # ),
        flow_rate=FlowRate.from_ul(20),
        delay=0.1,
    )
    await robot_api.motion.blow_out(
        pipette_number=1,
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.ready_plunger(
        pipette_number=1,
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.move_z_up()
    await robot_api.motion.move_to_ready()

    # deck_module.washer_dryer
    await robot_api.deck_module.washer_dryer.off_do()
    await robot_api.deck_module.washer_dryer.initialize()
    await robot_api.deck_module.washer_dryer.move_to_ready()
    await robot_api.deck_module.washer_dryer.move_to_washing()
    await robot_api.deck_module.washer_dryer.move_to(
        column=1,
        offset=(0.1, 0.2, 0.3),
    )
    await robot_api.deck_module.washer_dryer.prime(sec=5)
    await robot_api.deck_module.washer_dryer.recovery(sec=5)
    await robot_api.deck_module.washer_dryer.suction(
        columns=[1, 4, 7, 10],
        depth=10,
        z_speed=Speed.from_mm(5),
        delay_sec=0.1,
    )
    await robot_api.deck_module.washer_dryer.dispense(
        columns=[1, 4, 7, 10],
        volume=100,
    )
    await robot_api.deck_module.washer_dryer.wash_needle(
        depth=5,
        delay_sec=0.5,
    )
    await robot_api.deck_module.washer_dryer.wash_tube(
        cycle=3,
        suction_sec=1.0,
        dispense_sec=1.0,
    )
    await robot_api.deck_module.washer_dryer.wash(
        pre_dispense=True,
        columns=[1, 4, 7, 10],
        suction=True,
        suction_depth=9,
        suction_z_speed=10,
        suction_delay_sec=3,
        dispense=True,
        dispense_volume=100,
        wash_needle=True,
        wash_needle_depth=12,
        wash_needle_delay_sec=1,
    )
    await robot_api.deck_module.washer_dryer.dry(sec=3)

    # deck_module.magnetic_shaker
    await robot_api.deck_module.magnetic_shaker.initialize()
    await robot_api.deck_module.magnetic_shaker.magnet(on=True)
    await robot_api.deck_module.magnetic_shaker.magnet(on=False)
    await robot_api.deck_module.magnetic_shaker.shake(
        rpm=500,
        acceleration_sec=5,
    )
    await asyncio.sleep(5)
    await robot_api.deck_module.magnetic_shaker.shake_off()
    await robot_api.deck_module.magnetic_shaker.move_to_ready()

    # deck_module.shaker
    await robot_api.deck_module.shaker.initialize()
    await robot_api.deck_module.shaker.shake(
        rpm=500,
        acceleration_sec=5,
    )
    await asyncio.sleep(5)
    await robot_api.deck_module.shaker.shake_off()
    await robot_api.deck_module.shaker.move_to_ready()

    # axis
    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.jog(axis=Axis.X, value=0)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



import asyncio
from loguru import logger
import cv2

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.api.robot_api import RobotAPI as BaseRobotAPI
from ablelabs.neon.controllers.notable.kribb.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.kribb.api.set_api import SetAPI
from ablelabs.neon.controllers.notable.kribb.api.motion_api import MotionAPI
from ablelabs.neon.controllers.notable.kribb.api.optic_api import OpticAPI
from ablelabs.neon.common.notable.structs import Location


class RobotAPI(BaseRobotAPI):
    def __init__(self) -> None:
        tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)
        super().__init__(tcp_client)
        self._set_api = SetAPI(tcp_client=tcp_client)
        self._motion_api = MotionAPI(tcp_client=tcp_client)
        self._optic_api = OpticAPI(tcp_client=tcp_client)

    @property
    def set(self):
        return self._set_api

    @property
    def motion(self):
        return self._motion_api

    @property
    def optic(self):
        return self._optic_api

    @run_server_func(RobotRouter.robot_scan_displacement_sensor)
    async def scan_displacement_sensor(
        self,
        pipette_number: int,
        location: Location,
        padding: tuple[float, float],
        scan_count: tuple[float, float],
    ) -> tuple[list[float], list[float], list[list[float]], float, float, cv2.Mat]:
        pass

    @run_server_func(RobotRouter.robot_move_to_interpolated_z)
    async def move_to_interpolated_z(
        self,
        pipette_number: int,
        location: Location,
        x: float,
        y: float,
    ):
        pass



import os
import sys
from loguru import logger

REQUIRED_FILES = [
    # "deck.json",
    "region.json",
    "pipette.json",
    "tip.json",
    "labware.json",
    "setup_data.toml",
    "driver_param.toml",
]


class Path:
    def __init__(
        self,
        required_files: list[str] = REQUIRED_FILES,
        required_dirs: list[str] = [],
    ) -> None:
        is_production = Path.is_production()
        root_path = Path.get_root_path()
        # print(f"root_path = {root_path}")
        if is_production:
            config_path = Path.get_config_path(root_path, required_files)
        else:
            config_path = Path.get_config_path(root_path, required_files, required_dirs)
        self._config_path = config_path
        logger.debug(f"root_path={root_path}  config_path={config_path}")
        # sypark platform 의존성 주입에 따라 경로 설정 변경 필요.
        # self.DECK_PATH = os.path.join(config_path, "deck.json")
        self.REGION_PATH = os.path.join(config_path, "region.json")
        self.PIPETTE_PATH = os.path.join(config_path, "pipette.json")
        self.TIP_PATH = os.path.join(config_path, "tip.json")
        self.DECK_MODULE_PATH = os.path.join(config_path, "deck_module.json")
        self.LABWARE_PATH = os.path.join(config_path, "labware.json")
        self.SETUP_DATA_PATH = os.path.join(config_path, "setup_data.toml")
        self.DRIVER_PARAM_PATH = os.path.join(config_path, "driver_param.toml")
        
    @property
    def config_path(self):
        return self._config_path

    @staticmethod
    def is_production() -> bool:
        """PyInstaller에 의해 배포되었는지 여부"""
        # return getattr(sys, "_MEIPASS", False)
        return getattr(
            sys, "frozen", False
        )  # cx_Freeze 같은 다른 패키징 도구에도 해당.

    @staticmethod
    def get_root_path() -> str:
        if Path.is_production():
            root_path = os.path.dirname(
                os.path.abspath(sys.executable)
            )  # 실행 파일 경로.
        else:
            root_path = os.path.abspath(os.curdir)  # 프로젝트 경로.
        return root_path

    @staticmethod
    def get_config_path(
        root_path: str, required_files: list[str], required_dirs: list[str] = []
    ):
        for dirpath, dirnames, filenames in os.walk(root_path):
            if all(required_dir in dirpath for required_dir in required_dirs) and all(
                required_file in filenames for required_file in required_files
            ):
                return dirpath
        return None



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter
from ablelabs.neon.common.notable.nanophilia.structs import Speed


class WasherDryerAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.deck_module_washer_dryer_off_do)
    async def off_do(self):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_initialize)
    async def initialize(self):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_move_to_ready)
    async def move_to_ready(self):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_move_to_washing)
    async def move_to_washing(self):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_move_to)
    async def move_to(
        self,
        column: int,
        offset: tuple[float, float, float] = (0, 0, 0),
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_prime)
    async def prime(self, sec: float):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_recovery)
    async def recovery(self, sec: float):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_suction)
    async def suction(
        self,
        columns: list[int],
        depth: float,
        z_speed: Speed = None,
        delay_sec: float = 0,
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_dispense)
    async def dispense(
        self,
        columns: list[int],
        volume: float,
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_wash_needle)
    async def wash_needle(
        self,
        depth: float,
        delay_sec: float,
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_wash_tube)
    async def wash_tube(
        self,
        cycle: int,
        suction_sec: float,
        dispense_sec: float,
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_wash)
    async def wash(
        self,
        pre_dispense: bool,
        columns: list[int],
        suction: bool,
        suction_depth: float,
        suction_z_speed: float,
        suction_delay_sec: float,
        dispense: bool,
        dispense_volume: float,
        wash_needle: bool,
        wash_needle_depth: float,
        wash_needle_delay_sec: float,
    ):
        pass

    @run_server_func(RobotRouter.deck_module_washer_dryer_dry)
    async def dry(self, sec: float):
        pass



from ablelabs.neon_v2.notable.core import Base
from ablelabs.neon_v2.notable.enums import ZReference


class UpperModule(Base):
    # Main operation methods
    async def initialize(
        self, home_axes: bool = True, move_to_ready: bool = True
    ) -> None:
        """Initialize robot (homing all axes and move to ready position).

        Args:
            home_axes: Whether to perform homing on all axes
            move_to_ready: Whether to move to ready position after initialization
        """
        body = {"home_axes": home_axes, "move_to_ready": move_to_ready}
        return await self._post("/api/v1/controller/upper-module/initialize", body=body)

    async def move_z_up(self) -> None:
        """Move all Z axes to safe height."""
        return await self._post("/api/v1/controller/upper-module/move-z-up")

    async def move_to(
        self,
        pipette_number: int,
        deck_number: int,
        well: str = "A1",
        z_reference: ZReference = ZReference.TOP,
        xyz_offset: list[float, float, float] = None,
        z_speed: bool | float = False,
    ) -> None:
        """Move safely to specified deck well position.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            well: Well identifier (e.g., "A1", "B2")
            z_reference: Z-axis reference point from ZReference enum
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "well": well,
            "z_reference": z_reference,
            "z_speed": z_speed,
        }
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post("/api/v1/controller/upper-module/move-to", body=body)

    async def move_xy(self, x: float, y: float, safe_z_height: bool = True) -> None:
        """Move to XY coordinates after raising Z to safe height.

        Args:
            x: X coordinate in mm
            y: Y coordinate in mm
            safe_z_height: Whether to move to safe Z height first
        """
        body = {"x": x, "y": y, "safe_z_height": safe_z_height}
        return await self._post("/api/v1/controller/upper-module/move-xy", body=body)

    async def move_z(
        self,
        pipette_number: int,
        deck_number: int,
        z_reference: ZReference = ZReference.TOP,
        z_offset: float = 0,
        z_speed: bool | float = False,
    ) -> None:
        """Move only Z axis to specified height at current XY position.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            z_reference: Z-axis reference point from ZReference enum
            z_offset: Z-axis offset in mm (-100 to 100)
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "z_reference": z_reference,
            "z_offset": z_offset,
            "z_speed": z_speed,
        }
        return await self._post("/api/v1/controller/upper-module/move-z", body=body)

    async def step_z(
        self, pipette_number: int, z_offset: float, z_speed: bool | float = False
    ) -> None:
        """Move Z axis by relative distance from current position.

        Args:
            pipette_number: Pipette number (1-2)
            z_offset: Relative Z-axis offset in mm (-100 to 100)
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "z_offset": z_offset,
            "z_speed": z_speed,
        }
        return await self._post("/api/v1/controller/upper-module/step-z", body=body)

    async def pick_up_tip(
        self,
        pipette_number: int,
        deck_number: int,
        well: str = "A1",
        xyz_offset: list[float, float, float] = None,
    ) -> None:
        """Pick up tips at specified position.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            well: Well identifier (e.g., "A1", "B2")
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "well": well,
        }
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post(
            "/api/v1/controller/upper-module/pick-up-tip", body=body
        )

    async def drop_tip(
        self,
        pipette_number: int,
        deck_number: int = None,
        well: str = "A1",
        xyz_offset: list[float, float, float] = None,
    ) -> None:
        """Drop tips at specified position.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12), None for trash
            well: Well identifier (e.g., "A1", "B2")
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
        """
        body = {"pipette_number": pipette_number, "well": well}
        if deck_number is not None:
            body["deck_number"] = deck_number
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post("/api/v1/controller/upper-module/drop-tip", body=body)

    async def ready_plunger(
        self, pipette_number: int, flow_rate: bool | float = False
    ) -> None:
        """Move plunger to ready position for aspiration (first stop).
        
        This moves the plunger to the first stop position, which is the standard
        position for aspiration and dispense operations.
        
        IMPORTANT: This operation must be performed when the tip is NOT submerged
        in liquid to prevent liquid from being drawn into the pipette accidentally.

        Args:
            pipette_number: Pipette number (1-2)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
        """
        body = {"pipette_number": pipette_number, "flow_rate": flow_rate}
        return await self._post(
            "/api/v1/controller/upper-module/ready-plunger", body=body
        )

    async def blow_out(
        self, pipette_number: int, flow_rate: bool | float = False
    ) -> None:
        """Blow out remaining liquid in tip (second stop).
        
        This moves the plunger to the second stop position to expel any
        remaining liquid from the tip after dispensing.

        Args:
            pipette_number: Pipette number (1-2)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
        """
        body = {"pipette_number": pipette_number, "flow_rate": flow_rate}
        return await self._post("/api/v1/controller/upper-module/blow-out", body=body)

    async def drop_plunger(self, pipette_number: int) -> None:
        """Move plunger down for tip ejection.

        Args:
            pipette_number: Pipette number (1-2)
        """
        body = {"pipette_number": pipette_number}
        return await self._post(
            "/api/v1/controller/upper-module/drop-plunger", body=body
        )

    async def aspirate(
        self,
        pipette_number: int,
        volume: float | list[float],
        flow_rate: bool | float,
        z_offset: float = 0,
        pause_sec: float = 0,
    ) -> None:
        """Aspirate specified volume of liquid.

        Args:
            pipette_number: Pipette number (1-2)
            volume: Volume to aspirate in μL (>0, ≤200). Can be single value or list for multi-aspirate
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            z_offset: Z-axis offset in mm (-100 to 100)
            pause_sec: Pause duration after aspiration in seconds (0-600)
            
        Example:
            # Single aspirate
            await upper_module.aspirate(1, 100.0, 50.0, z_offset=1.0, pause_sec=0.5)
            
            # Multi-aspirate for 8-channel pipette
            await upper_module.aspirate(1, [100.0, 150.0, 200.0], True)
        """
        body = {
            "pipette_number": pipette_number,
            "volume": volume,
            "flow_rate": flow_rate,
            "z_offset": z_offset,
            "pause_sec": pause_sec,
        }
        return await self._post("/api/v1/controller/upper-module/aspirate", body=body)

    async def dispense(
        self,
        pipette_number: int,
        volume: float,
        flow_rate: bool | float,
        z_offset: float = 0,
        pause_sec: float = 0,
    ) -> None:
        """Dispense specified volume of liquid.

        Args:
            pipette_number: Pipette number (1-2)
            volume: Volume to dispense in μL (0-200)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            z_offset: Z-axis offset in mm (-100 to 100)
            pause_sec: Pause duration after dispensing in seconds (0-600)
        """
        body = {
            "pipette_number": pipette_number,
            "volume": volume,
            "flow_rate": flow_rate,
            "z_offset": z_offset,
            "pause_sec": pause_sec,
        }
        return await self._post("/api/v1/controller/upper-module/dispense", body=body)

    async def mix(
        self,
        pipette_number: int,
        cycle: int,
        volume: float,
        flow_rate: bool | float,
        delay: float = 0,
        z_offset: float = 0,
    ) -> None:
        """Mix liquid at current position.

        Args:
            pipette_number: Pipette number (1-2)
            cycle: Number of mixing cycles (1-100)
            volume: Volume to mix in μL (0-200)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            delay: Delay between cycles in seconds (0-10)
            z_offset: Z-axis offset in mm (-100 to 100)
        """
        body = {
            "pipette_number": pipette_number,
            "cycle": cycle,
            "volume": volume,
            "flow_rate": flow_rate,
            "delay": delay,
            "z_offset": z_offset,
        }
        return await self._post("/api/v1/controller/upper-module/mix", body=body)

    # Estimation methods
    async def estimate_initialize(
        self, home_axes: bool = True, move_to_ready: bool = True
    ) -> None:
        """Estimate time for initialization.

        Args:
            home_axes: Whether to perform homing on all axes
            move_to_ready: Whether to move to ready position after initialization
        """
        body = {"home_axes": home_axes, "move_to_ready": move_to_ready}
        return await self._post(
            "/api/v1/controller/upper-module/estimate/initialize", body=body
        )

    async def estimate_move_z_up(self) -> None:
        """Estimate time for moving all Z axes up."""
        return await self._post("/api/v1/controller/upper-module/estimate/move-z-up")

    async def estimate_move_to(
        self,
        pipette_number: int,
        deck_number: int,
        well: str = "A1",
        z_reference: ZReference = ZReference.TOP,
        xyz_offset: list[float, float, float] = None,
        z_speed: bool | float = False,
    ) -> None:
        """Estimate time for move_to operation.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            well: Well identifier (e.g., "A1", "B2")
            z_reference: Z-axis reference point from ZReference enum
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "well": well,
            "z_reference": z_reference,
            "z_speed": z_speed,
        }
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post(
            "/api/v1/controller/upper-module/estimate/move-to", body=body
        )

    async def estimate_move_xy(
        self, x: float, y: float, safe_z_height: bool = True
    ) -> None:
        """Estimate time for XY movement.

        Args:
            x: X coordinate in mm
            y: Y coordinate in mm
            safe_z_height: Whether to move to safe Z height first
        """
        body = {"x": x, "y": y, "safe_z_height": safe_z_height}
        return await self._post(
            "/api/v1/controller/upper-module/estimate/move-xy", body=body
        )

    async def estimate_move_z(
        self,
        pipette_number: int,
        deck_number: int,
        z_reference: ZReference = ZReference.TOP,
        z_offset: float = 0,
        z_speed: bool | float = False,
    ) -> None:
        """Estimate time for Z movement.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            z_reference: Z-axis reference point from ZReference enum
            z_offset: Z-axis offset in mm (-100 to 100)
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "z_reference": z_reference,
            "z_offset": z_offset,
            "z_speed": z_speed,
        }
        return await self._post(
            "/api/v1/controller/upper-module/estimate/move-z", body=body
        )

    async def estimate_step_z(
        self, pipette_number: int, z_offset: float, z_speed: bool | float = False
    ) -> None:
        """Estimate time for Z step movement.

        Args:
            pipette_number: Pipette number (1-2)
            z_offset: Relative Z-axis offset in mm (-100 to 100)
            z_speed: Z-axis speed in mm/s (float) or automatic (True=max, False=no change)
        """
        body = {
            "pipette_number": pipette_number,
            "z_offset": z_offset,
            "z_speed": z_speed,
        }
        return await self._post(
            "/api/v1/controller/upper-module/estimate/step-z", body=body
        )

    async def estimate_pick_up_tip(
        self,
        pipette_number: int,
        deck_number: int,
        well: str = "A1",
        xyz_offset: list[float, float, float] = None,
    ) -> None:
        """Estimate time for pick_up_tip operation.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12)
            well: Well identifier (e.g., "A1", "B2")
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
        """
        body = {
            "pipette_number": pipette_number,
            "deck_number": deck_number,
            "well": well,
        }
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post(
            "/api/v1/controller/upper-module/estimate/pick-up-tip", body=body
        )

    async def estimate_drop_tip(
        self,
        pipette_number: int,
        deck_number: int = None,
        well: str = "A1",
        xyz_offset: list[float, float, float] = None,
    ) -> None:
        """Estimate time for drop_tip operation.

        Args:
            pipette_number: Pipette number (1-2)
            deck_number: Deck position number (1-12), None for trash
            well: Well identifier (e.g., "A1", "B2")
            xyz_offset: XYZ coordinate offset in mm [X, Y, Z]
        """
        body = {"pipette_number": pipette_number, "well": well}
        if deck_number is not None:
            body["deck_number"] = deck_number
        if xyz_offset is not None:
            body["xyz_offset"] = xyz_offset
        return await self._post(
            "/api/v1/controller/upper-module/estimate/drop-tip", body=body
        )

    async def estimate_ready_plunger(
        self, pipette_number: int, flow_rate: bool | float = False
    ) -> None:
        """Estimate time for plunger ready operation (first stop).
        
        Calculates the estimated time to move plunger to first stop position.
        
        IMPORTANT: This operation must be performed when the tip is NOT submerged
        in liquid to prevent liquid from being drawn into the pipette accidentally.

        Args:
            pipette_number: Pipette number (1-2)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
        """
        body = {"pipette_number": pipette_number, "flow_rate": flow_rate}
        return await self._post(
            "/api/v1/controller/upper-module/estimate/ready-plunger", body=body
        )

    async def estimate_blow_out(
        self, pipette_number: int, flow_rate: bool | float = False
    ) -> None:
        """Estimate time for blow out operation (second stop).
        
        Calculates the estimated time to move plunger to second stop position.

        Args:
            pipette_number: Pipette number (1-2)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
        """
        body = {"pipette_number": pipette_number, "flow_rate": flow_rate}
        return await self._post(
            "/api/v1/controller/upper-module/estimate/blow-out", body=body
        )

    async def estimate_drop_plunger(self, pipette_number: int) -> None:
        """Estimate time for plunger drop operation.

        Args:
            pipette_number: Pipette number (1-2)
        """
        body = {"pipette_number": pipette_number}
        return await self._post(
            "/api/v1/controller/upper-module/estimate/drop-plunger", body=body
        )

    async def estimate_aspirate(
        self,
        pipette_number: int,
        volume: float,
        flow_rate: bool | float,
        z_offset: float = 0,
        pause_sec: float = 0,
    ) -> None:
        """Estimate time for aspirate operation.

        Args:
            pipette_number: Pipette number (1-2)
            volume: Volume to aspirate in μL (0-200)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            z_offset: Z-axis offset in mm (-100 to 100)
            pause_sec: Pause duration after aspiration in seconds (0-600)
        """
        body = {
            "pipette_number": pipette_number,
            "volume": volume,
            "flow_rate": flow_rate,
            "z_offset": z_offset,
            "pause_sec": pause_sec,
        }
        return await self._post(
            "/api/v1/controller/upper-module/estimate/aspirate", body=body
        )

    async def estimate_dispense(
        self,
        pipette_number: int,
        volume: float,
        flow_rate: bool | float,
        z_offset: float = 0,
        pause_sec: float = 0,
    ) -> None:
        """Estimate time for dispense operation.

        Args:
            pipette_number: Pipette number (1-2)
            volume: Volume to dispense in μL (0-200)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            z_offset: Z-axis offset in mm (-100 to 100)
            pause_sec: Pause duration after dispensing in seconds (0-600)
        """
        body = {
            "pipette_number": pipette_number,
            "volume": volume,
            "flow_rate": flow_rate,
            "z_offset": z_offset,
            "pause_sec": pause_sec,
        }
        return await self._post(
            "/api/v1/controller/upper-module/estimate/dispense", body=body
        )

    async def estimate_mix(
        self,
        pipette_number: int,
        cycle: int,
        volume: float,
        flow_rate: bool | float,
        delay: float = 0,
        z_offset: float = 0,
    ) -> None:
        """Estimate time for mix operation.

        Args:
            pipette_number: Pipette number (1-2)
            cycle: Number of mixing cycles (1-100)
            volume: Volume to mix in μL (0-200)
            flow_rate: Flow rate in μL/s (0-500) or automatic (True=max, False=no change)
            delay: Delay between cycles in seconds (0-10)
            z_offset: Z-axis offset in mm (-100 to 100)
        """
        body = {
            "pipette_number": pipette_number,
            "cycle": cycle,
            "volume": volume,
            "flow_rate": flow_rate,
            "delay": delay,
            "z_offset": z_offset,
        }
        return await self._post(
            "/api/v1/controller/upper-module/estimate/mix", body=body
        )



# python 3.10.0

# Install the ablelabs package using "pip install ablelabs"
# Import necessary modules from ablelabs package
from ablelabs.neon.controllers.notable.api.robot_api import RobotAPI
from ablelabs.neon.common.notable.enums import (
    Axis,
    LocationReference,
    RunStatus,
)
from ablelabs.neon.common.notable.structs import Speed, FlowRate, location


async def main():
    import subprocess

    # Launch Neon.exe using subprocess (for Windows systems)
    # Ensure 'Neon.exe' and the 'resources' folder are in the correct directory
    subprocess.Popen(
        [
            r"C:\your_path\Neon.exe",
        ]
    )

    # Instantiate the robot API client
    robot_api = RobotAPI()
    # Ensure to follow the sequence: connect -> wait_boot -> set_pipettes/tips/labwares -> initialize -> motion_api for proper operation

    # Connect to the robot system
    ip = "localhost"
    port = 1234
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass  # Handle exceptions for connection failures

    # Wait for the robot system to boot
    await robot_api.wait_boot()
    # LED bar indicates the robot's current run_status
    # READY (white), INITIALIZE (blue), RUN (green), PAUSE (yellow), ERROR (red)
    await robot_api.set_run_status(RunStatus.READY)
    await robot_api.get_run_status()

    # --- Set up robot components ---

    # Configure pipettes
    await robot_api.set.pipettes(
        {
            # 1: "8ch1000ul",
            2: "1ch200ul",
        }
    )

    # Configure tips
    await robot_api.set.tips(
        {
            # 1: "tip_1000",
            2: "tip_200",
        }
    )

    # Configure labware layout on the deck
    await robot_api.set.labwares(
        {
            1: "spl_trayplate_60ml_#30001",
            2: "spl_96_well_0.2ml_#30096",
            10: "ablelabs_tiprack_#AL-CT-200",
            11: "ablelabs_tiprack_#AL-CT-1000",
            12: "trash_#v2.5",
        }
    )

    # Retrieve robot setup data
    await robot_api.get.pipette_infos()
    await robot_api.get.tip_infos()
    await robot_api.get.deck_module_infos()
    await robot_api.get.labware_infos()
    setup_data = await robot_api.get.setup_data()

    # Modify setup data: Example for teaching offsets
    x, y, z = setup_data["teaching"]["pipette_1"]["8ch200ul"]["deck1_up"]
    setup_data["teaching"]["pipette_1"]["8ch200ul"]["deck1_up"] = [x + 1, y + 2, z + 3]
    await robot_api.set.update_setup_data(setup_data, True)

    # --- Robot Control Functions ---

    # Stop, clear errors, pause, and resume robot operation
    await robot_api.stop()
    await robot_api.clear_error()  # Clear errors after stopping
    await robot_api.pause()
    await robot_api.resume()

    # Check if the robot is still connected
    is_connected = await robot_api.is_connected()

    # --- Motion Commands ---

    # Initialize the motion system
    await robot_api.motion.initialize()

    # Move pipette to a specified location with an offset
    await robot_api.motion.move_to(
        pipette_number=2,
        location=location(
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(1.0, 2.0, 3.0),
        ),
    )

    # Pick up and drop tips at specified locations
    await robot_api.motion.pick_up_tip(
        pipette_number=2,
        location=location(
            location_number=10,
            well="a1",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=2,
        location=location(
            location_number=10,
            well="a1",
        ),
    )

    # Aspirate and dispense liquids at specified locations
    await robot_api.motion.aspirate(
        pipette_number=2,
        volume=200,
        location=location(
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.motion.dispense(
        pipette_number=2,
        volume=200,
        location=location(
            location_number=2,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )

    # Mix liquids at the last specified position
    await robot_api.motion.mix(
        pipette_number=2,
        volume=100,
        iteration=2,
        flow_rate=FlowRate.from_ul(70),
        delay=0.1,
    )

    # Blow-out liquids to clear the pipette tip
    await robot_api.motion.blow_out(
        pipette_number=2,
        flow_rate=FlowRate.from_ul(200),
    )

    # Return pipette to the ready position
    await robot_api.motion.move_to_ready()

    # --- Axis Control Commands ---

    # Retrieve, set, and adjust axis parameters (X-axis example)
    position = await robot_api.axis.get_position(axis=Axis.X)  # Get position in mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # Set speed in mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # Set acceleration in mm/sec2
    await robot_api.axis.disable(axis=Axis.X)  # Disable the axis
    await robot_api.axis.enable(axis=Axis.X)  # Enable the axis
    await robot_api.axis.stop(axis=Axis.X)  # Stop the axis movement
    await robot_api.axis.home(axis=Axis.X)  # Perform homing operation
    await robot_api.axis.wait_home_done(axis=Axis.X)  # Wait for homing to complete
    await robot_api.axis.move(axis=Axis.X, value=10)  # Move axis by 10mm
    await robot_api.axis.wait_move_done(axis=Axis.X)  # Wait for movement to complete


if __name__ == "__main__":
    # Run the main async function with an event loop
    import asyncio

    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.suitable.enums import Axis

DI_COUNT: int = 8
DO_COUNT: int = 8

PIPETTE_COUNT: int = 8
PIPETTE_NUMBERS = [i + 1 for i in range(PIPETTE_COUNT)]

AXIS_YS = [Axis.Y[n] for n in PIPETTE_NUMBERS]
AXIS_ZS = [Axis.Z[n] for n in PIPETTE_NUMBERS]
AXIS_PS = [Axis.P[n] for n in PIPETTE_NUMBERS]

Y_MIN_INTERVAL_MM = 9.0



from ablelabs.neon_v2.notable.core import Base


class Pipette(Base):
    """Pipette configuration management for liquid handling setup.
    
    This class manages pipette configuration for the robot's dual-pipette system.
    Each pipette mount (1 and 2) can be configured with different pipette types
    based on the experimental requirements.
    """

    async def get_pipette(self) -> dict:
        """Get current pipette configuration.
        
        Retrieves the currently configured pipette types for both pipette mounts.
        Each mount can have a different pipette type or be empty.
        
        Returns:
            dict: Current pipette configuration mapping mount positions to pipette types.
                Format: {"1": "pipette_type", "2": "another_pipette"}
                Empty mounts are not included in the response.
                
        Example:
            {
                "1": "8ch_200ul",    # 8-channel 200μL pipette on mount 1
                "2": "1ch_1000ul"    # 1-channel 1000μL pipette on mount 2
            }
        """
        return await self._get("/api/v1/config/pipette/")

    async def set_pipette(self, config: dict[str | int, str]) -> dict:
        """Set pipette configuration for liquid handling.
        
        Configures which pipette codes are mounted on each pipette position.
        This configuration determines the liquid handling capabilities and
        volume ranges available for protocols.
        
        Args:
            config: Dictionary mapping pipette mount positions to pipette codes.
                   Positions can be strings ("1", "2") or integers (1, 2).
                   Values must be valid pipette codes from the pipette library.
                   
        Returns:
            dict: Complete pipette configuration with all positions.
                 Format: {"1": "pipette_code", "2": "pipette_code"}
                 Unassigned positions return None.
            
        Example:
            config = {
                1: "8ch_200ul",       # 8-channel 200μL pipette on mount 1 (int key)
                "2": "1ch_200ul"      # 1-channel 200μL pipette on mount 2 (str key)  
            }
            
            # Returns both positions:
            {
                "1": "8ch_200ul",
                "2": "1ch_200ul"
            }
            
        Available Pipette Codes:
            - "8ch_200ul": 8-channel 200μL pipette (multi-channel)
            - "8ch_1000ul": 8-channel 1000μL pipette (multi-channel)
            - "1ch_200ul": 1-channel 200μL pipette (single-channel)
            - "1ch_20ul": 1-channel 20μL pipette (precision)
            
        Note:
            - Position keys can be strings or integers (both accepted)
            - Positions must be 1 or 2 (dual pipette system)
            - Pipette codes must exist in the pipette library
            - Invalid pipette codes will cause API errors
            - Empty mounts should be omitted from input (will return None)
            - Multi-channel pipettes require compatible labware (96-well plates, etc.)
        """
        return await self._post("/api/v1/config/pipette/", body=config)


# python 3.10+

# Install the ablelabs package using "pip install ablelabs"
import asyncio
import subprocess
from ablelabs.neon_v2.notable import Notable, LEDColor, ZReference

NEON_EXE_PATH = r"C:\your_path\NOTABLE-Neon.exe"


async def launch_neon_server(exe_path: str = NEON_EXE_PATH, wait_seconds: int = 5):
    """
    Launch NOTABLE-Neon.exe server and wait for initialization.

    Prerequisites:
    1. Place 'resources' folder in the same directory as NOTABLE-Neon.exe
    2. Check COM port in Device Manager and verify resources/setup_data.toml [communication.port]
    3. Configure [env] is_online setting in setup_data.toml:
       - is_online = true: Connect to actual hardware (production mode)
       - is_online = false: Run server without hardware connection (simulation mode)

    Alternative execution methods:
    - Double-click NOTABLE-Neon.exe manually
    - Run via command line: NOTABLE-Neon.exe
    - Use Windows service or task scheduler

    The server provides:
    - REST API endpoints at http://localhost:7777
    - Interactive Swagger UI documentation at http://localhost:7777/docs
    - OpenAPI specification at http://localhost:7777/openapi.json
    - ReDoc documentation at http://localhost:7777/redoc

    Args:
        exe_path: Path to NOTABLE-Neon.exe (None for default path)
        wait_seconds: Seconds to wait for server initialization
    """
    print("Launching NOTABLE-Neon.exe server...")
    print("📁 Ensure 'resources' folder is in the same directory as the executable")
    print("🔌 Check COM port settings in Device Manager and setup_data.toml")
    print("🌐 Configure [env] is_online in setup_data.toml:")
    print("   - is_online = true: Production mode (connects to actual hardware)")
    print("   - is_online = false: Simulation mode (no hardware connection)")
    print("📖 Access API documentation at http://localhost:7777/docs after launch")

    try:
        # Launch Neon.exe server
        subprocess.Popen([exe_path])
        print(f"⏳ Waiting {wait_seconds} seconds for server initialization...")
        await asyncio.sleep(wait_seconds)
        print("✅ Server should now be running at http://localhost:7777")

    except FileNotFoundError:
        print(f"❌ NOTABLE-Neon.exe not found at: {exe_path}")
        print(
            "💡 Alternative: Launch NOTABLE-Neon.exe manually before running this script"
        )
        raise
    except Exception as e:
        print(f"❌ Failed to launch server: {e}")
        print("💡 Try launching NOTABLE-Neon.exe manually")
        raise


async def setup_notable_client(base_url: str = "http://localhost:7777"):
    """
    Create Notable API client and retrieve library information.

    Args:
        base_url: Notable server URL

    Returns:
        tuple: (notable_client, pipette_library, labware_library)
    """
    print("Setting up Notable API client...")
    notable = Notable(base_url)

    # Get available resources
    print("📚 Retrieving pipette and labware libraries...")
    pipette_library = await notable.resource.library.get_pipette()
    pipette_codes = list(pipette_library.keys())
    labware_library = await notable.resource.library.get_labware()
    labware_codes = {
        category: list(dict(labware_codes).keys())
        for category, labware_codes in labware_library.items()
    }

    print(f"Available pipette codes: {pipette_codes}")
    print(f"Available labware codes: {labware_codes}")

    return notable, pipette_library, labware_library


async def simple_example():
    """Simple liquid handling workflow - minimal setup for quick start."""
    print("=== Simple Version: Basic Liquid Handling ===")

    # Setup client
    notable, pipette_library, labware_library = await setup_notable_client()

    # Set LED to indicate ready
    await notable.driver.io.set_led_bar(
        color=LEDColor.GREEN,
        bright_percent=20,
        progress_percent=100,
        blink_time_ms=0,  # No blinking
    )

    # Configure pipettes and deck layout
    await notable.config.pipette.set_pipette(
        {
            "1": "8ch_200ul",
        }
    )
    await notable.config.deck.set_deck(
        {
            "9": "ablelabs_tip_box_200",  # Tip box
            "6": "nest_12_reservoir_360102",  # Reservoir
            "3": "spl_96_well_plate_30096",  # 96-well plate
            "12": "ablelabs_trash",  # Trash
        }
    )

    # Basic liquid handling workflow
    await notable.controller.upper_module.initialize()
    await notable.controller.upper_module.pick_up_tip(
        pipette_number=1,
        deck_number="9",
        well="A1",
    )

    # Aspirate from reservoir
    await notable.controller.upper_module.move_to(
        pipette_number=1,
        deck_number="6",
        well="A1",
        z_reference=ZReference.BOTTOM,
    )
    await notable.controller.upper_module.aspirate(
        pipette_number=1,
        volume=100.0,
    )

    # Dispense to plate
    await notable.controller.upper_module.move_to(
        pipette_number=1,
        deck_number="3",
        well="A1",
        z_reference=ZReference.BOTTOM,
    )
    await notable.controller.upper_module.dispense(
        pipette_number=1,
        volume=100.0,
    )
    await notable.controller.upper_module.blow_out(
        pipette_number=1,
    )

    await notable.controller.upper_module.drop_tip(
        pipette_number=1,
        deck_number="12",
    )

    print("✓ Simple liquid handling completed!")


async def complex_example():
    """Complex liquid handling workflow - multiple transfers with different volumes."""
    print("=== Complex Version: Multiple Transfers ===")

    # Setup client
    notable, pipette_library, labware_library = await setup_notable_client()

    # Set LED to indicate ready
    await notable.driver.io.set_led_bar(
        color=LEDColor.WHITE,
        bright_percent=20,
        progress_percent=100,
        blink_time_ms=0,  # No blinking
    )

    pipette_number = 1
    pipette_code = "8ch_200ul"
    trash_deck_number = 12
    tip_deck_number = 9
    tip_labware_code = "ablelabs_tip_box_200"
    source_deck_number = 6
    target_deck_number = 3
    target_wells = [f"A{column}" for column in range(1, 13)]
    volume = 20.0
    flow_rate = 100
    withdraw_tip_z_speed = 20

    # Configure pipettes and deck layout
    await notable.config.pipette.set_pipette(
        {
            pipette_number: pipette_code,
        }
    )
    await notable.config.deck.set_deck(
        {
            tip_deck_number: tip_labware_code,  # Tip box
            source_deck_number: "nest_12_reservoir_360102",  # Reservoir
            target_deck_number: "spl_96_well_plate_30096",  # 96-well plate
            trash_deck_number: "ablelabs_trash",  # Trash
        }
    )

    def _chunk_list(lst: list[str], chunk_size: int) -> list[list[str]]:
        """Split a list into chunks of specified size."""
        return [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]

    # Calculate optimal chunk size for multi-dispense operations
    # Use minimum of pipette working volume and tip capacity to determine max volume per aspirate
    pipette = pipette_library[pipette_code]
    tip_labware = labware_library["tip_box"][tip_labware_code]
    tip = tip_labware["tip_info"]
    max_volume = min(pipette["working_volume"], tip["volume"])  # Safety constraint
    chunk_size = int(
        max_volume // volume
    )  # Number of wells that can be dispensed per aspirate
    chunked_wells = _chunk_list(
        target_wells, chunk_size
    )  # Group wells for batch processing

    print(
        ", ".join(
            [
                f"Complex Version: Multi-Dispense",
                f"pipette=({pipette_number}, {pipette_code})",
                f"tip=({tip_deck_number}, {tip_labware_code})",
                f"source={source_deck_number}",
                f"target={target_deck_number}",
                f"target_wells={target_wells}",
                f"volume={volume}",
                f"chunk_size={chunk_size}",
                f"chunked_wells={chunked_wells}",
            ]
        )
    )

    # Initialize robot and start liquid handling workflow
    await notable.driver.io.set_led_bar(
        color=LEDColor.BLUE,
        blink_time_ms=1000,  # Indicate initialization in progress
    )
    await notable.controller.upper_module.initialize(
        move_to_ready=False,  # Don't move to ready position to save time
    )
    await notable.driver.io.set_led_bar(
        color=LEDColor.GREEN,
        blink_time_ms=0,
        progress_percent=0,
    )

    # Pick up tips from the first position
    await notable.controller.upper_module.pick_up_tip(
        pipette_number=pipette_number,
        deck_number=tip_deck_number,
        well="A1",
    )

    # Update progress indicator (tip pickup counts as first step)
    n = len(chunked_wells) + 1  # +1 for tip pickup step
    await notable.driver.io.set_led_bar(
        progress_percent=int(100 * 1 / n),
    )

    # Process each chunk of wells
    for i, wells in enumerate(chunked_wells):
        print(
            f"--- Transfer {i+1}/{len(chunked_wells)}: {volume}μL to {len(wells)} wells ---"
        )

        # Pre-wet tip on first iteration for accuracy
        if i == 0:
            await notable.controller.upper_module.move_to(
                pipette_number=pipette_number,
                deck_number=source_deck_number,
                well="A1",
                z_reference=ZReference.BOTTOM,
                z_speed=True,
            )
            await notable.controller.upper_module.mix(
                pipette_number=pipette_number,
                cycle=3,
                volume=[volume] * len(wells),
                flow_rate=flow_rate,
            )
            await notable.controller.upper_module.blow_out(
                pipette_number=pipette_number,
            )
            await notable.controller.upper_module.move_z(
                pipette_number=pipette_number,
                deck_number=source_deck_number,
                z_reference=ZReference.TOP,
                z_speed=withdraw_tip_z_speed,
            )
            await notable.controller.upper_module.ready_plunger(
                pipette_number=pipette_number,
            )

        # Aspirate liquid from source reservoir
        await notable.controller.upper_module.move_to(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            well="A1",
            z_reference=ZReference.BOTTOM,
            z_speed=True,
        )
        await notable.controller.upper_module.aspirate(
            pipette_number=pipette_number,
            volume=[volume]
            * len(wells),  # aspirate for multi-well dispense: one volume per well
            flow_rate=flow_rate,
            pause_sec=0.5,  # Pause for liquid settling
        )
        await notable.controller.upper_module.move_z(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            z_reference=ZReference.TOP_JUST,
            z_speed=withdraw_tip_z_speed,
        )

        # Dispense to target wells sequentially
        for well in wells:
            await notable.controller.upper_module.move_to(
                pipette_number=pipette_number,
                deck_number=target_deck_number,
                well=well,
                z_reference=ZReference.BOTTOM,
                xyz_offset=[0, 0, 2],
                z_speed=True,
            )
            await notable.controller.upper_module.dispense(
                pipette_number=pipette_number,
                volume=volume,  # Single well dispense
                flow_rate=flow_rate,
                pause_sec=0.5,  # Pause for complete dispensing
            )
            await notable.controller.upper_module.move_z(
                pipette_number=pipette_number,
                deck_number=target_deck_number,
                z_reference=ZReference.TOP_JUST,
                z_speed=withdraw_tip_z_speed,
            )

        # Blow out any remaining liquid back to source reservoir
        await notable.controller.upper_module.move_to(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            well="A1",
            z_reference=ZReference.BOTTOM,
            xyz_offset=[0, 0, 2],
            z_speed=True,
        )
        await notable.controller.upper_module.blow_out(
            pipette_number=pipette_number,
        )
        await notable.controller.upper_module.move_z(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            z_reference=ZReference.TOP_JUST,
            z_speed=withdraw_tip_z_speed,
        )
        # Ready plunger for next iteration (if any)
        await notable.controller.upper_module.ready_plunger(
            pipette_number=pipette_number,
        )

        # Update progress indicator
        await notable.driver.io.set_led_bar(
            progress_percent=int(100 * (i + 2) / n),
        )

    # Drop tips to trash when all transfers are complete
    await notable.controller.upper_module.drop_tip(
        pipette_number=pipette_number,
    )

    # Final LED indication
    await notable.driver.io.set_led_bar(
        color=LEDColor.WHITE,
        progress_percent=100,
    )

    print("✓ Complex liquid handling completed!")


async def serial_dilution_example():
    """Serial dilution example."""
    print("=== Serial Dilution Version ===")

    # Setup client
    notable, pipette_library, labware_library = await setup_notable_client()

    pipette_number = 1
    pipette_code = "8ch_200ul"
    trash_deck_number = 12
    tip_deck_number = 1
    tip_labware_code = "ablelabs_tip_box_200"
    reservoir_deck_number = 2
    well_plate_deck_number = 3
    flow_rate = 100
    withdraw_tip_z_speed = 20

    tip_index = 1

    def _get_tip_well() -> str:
        return f"A{tip_index}"

    async def _transfer(
        source_deck_well: tuple[int, str],
        target_deck_well: tuple[int, str],
        volume: float,
        target_mix_count: int = 0,
    ):
        source_deck_number, source_well = source_deck_well
        target_deck_number, target_well = target_deck_well

        # aspirate from source
        await notable.controller.upper_module.move_to(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            well=source_well,
            z_reference=ZReference.BOTTOM,
        )
        await notable.controller.upper_module.aspirate(
            pipette_number=pipette_number,
            volume=volume,
            flow_rate=flow_rate,
        )
        await notable.controller.upper_module.move_z(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            z_reference=ZReference.TOP_JUST,
            z_speed=withdraw_tip_z_speed,
        )

        # dispense & mix to target
        await notable.controller.upper_module.move_to(
            pipette_number=pipette_number,
            deck_number=target_deck_number,
            well=target_well,
            z_reference=ZReference.BOTTOM,
        )
        await notable.controller.upper_module.dispense(
            pipette_number=pipette_number,
            volume=volume,
            flow_rate=flow_rate,
        )
        if target_mix_count > 0:
            await notable.controller.upper_module.mix(
                pipette_number=pipette_number,
                cycle=target_mix_count,
                volume=volume,
                flow_rate=flow_rate,
            )
        await notable.controller.upper_module.blow_out(
            pipette_number=pipette_number,
        )
        await notable.controller.upper_module.move_z(
            pipette_number=pipette_number,
            deck_number=source_deck_number,
            z_reference=ZReference.TOP_JUST,
            z_speed=withdraw_tip_z_speed,
        )
        await notable.controller.upper_module.ready_plunger(
            pipette_number=pipette_number,
        )

    # pipette = pipette_library[pipette_code]
    # tip_labware = labware_library["tip_box"][tip_labware_code]
    # tip = tip_labware["tip_info"]
    # max_volume = min(pipette["working_volume"], tip["volume"])  # Safety constraint

    await notable.config.pipette.set_pipette(
        {
            pipette_number: pipette_code,
        }
    )
    await notable.config.deck.set_deck(
        {
            tip_deck_number: tip_labware_code,  # Tip box
            reservoir_deck_number: "nest_12_reservoir_360102",  # Reservoir
            well_plate_deck_number: "spl_96_well_plate_30096",  # 96-well plate
            trash_deck_number: "ablelabs_trash",  # Trash
        }
    )

    await notable.controller.upper_module.initialize(
        move_to_ready=False,  # Don't move to ready position to save time
    )

    await notable.controller.upper_module.pick_up_tip(
        pipette_number=pipette_number,
        deck_number=tip_deck_number,
        well=_get_tip_well(),
    )

    for target_well in ["A2", "A3", "A4", "A5", "A6"]:
        await _transfer(
            (reservoir_deck_number, "A1"),
            (well_plate_deck_number, target_well),
            volume=120.0,
        )

    await notable.controller.upper_module.drop_tip(
        pipette_number=pipette_number,
    )
    tip_index += 1
    await notable.controller.upper_module.pick_up_tip(
        pipette_number=pipette_number,
        deck_number=tip_deck_number,
        well=_get_tip_well(),
    )

    for source_well, target_well in zip(
        ["A1", "A2", "A3", "A4", "A5"],
        ["A2", "A3", "A4", "A5", "A6"],
    ):
        await _transfer(
            (well_plate_deck_number, source_well),
            (well_plate_deck_number, target_well),
            volume=60.0,
            target_mix_count=3,
        )

    await notable.controller.upper_module.drop_tip(
        pipette_number=pipette_number,
    )

    print("✓ Serial dilution completed!")


async def main():
    """
    Main function to run both simple and complex examples.

    You can modify this to run only the example you need:
    - await simple_example()     # Basic liquid handling
    - await complex_example()    # Multi-well transfer with optimization
    - await serial_dilution_example()    # Liquid handling applications: Serial dilution
    """
    print("NOTABLE Robot Liquid Handling Examples")
    print("=" * 40)

    # Launch server once for both examples (efficiency)
    await launch_neon_server()

    # Run examples sequentially
    await simple_example()
    await complex_example()
    await serial_dilution_example()

    print("\n🎉 All examples completed successfully!")


if __name__ == "__main__":
    asyncio.run(main())



from ablelabs.neon_v2.notable.core import Base


class Robot(Base):
    """Robot action control for protocol execution management.
    
    This class provides methods to control robot protocol execution,
    including status monitoring, pause/resume functionality, and emergency stops.
    """

    async def get_status(self) -> dict:
        """Get current robot status.
        
        Returns:
            dict: Robot status information with keys:
                - status: Current robot state (BOOT, CONNECTED, INITIALIZING, READY, RUNNING, EMERGENCY, STOPPED, ERROR)
                - is_paused: Whether robot is paused
                - error_message: Error description (None if no error)
                - is_operational: Whether robot can perform operations
                - is_error: Whether robot is in error state
                - valid_transitions: Available next states
                
        Key States:
            - BOOT: System startup, hardware communication not established
            - CONNECTED: Hardware connected but requires initialization before use
            
        Note: Other states (INITIALIZING, READY, RUNNING, EMERGENCY, STOPPED, ERROR) 
        may not transition correctly in current implementation - use as reference only.
        """
        return await self._get("/api/v1/action/robot/status")

    async def pause(self) -> None:
        """Pause robot operation.
        
        Pauses all robot operations and prevents execution of subsequent commands.
        The robot completes the current operation before pausing.
        
        Raises:
            APIError: If robot is in error state or cannot be paused
        """
        return await self._post("/api/v1/action/robot/status/pause")

    async def resume(self) -> None:
        """Resume robot operation.
        
        Resumes robot operation from the paused state.
        
        Raises:
            APIError: If robot is not in paused state
        """
        return await self._post("/api/v1/action/robot/status/resume")

    async def stop(self) -> None:
        """Stop robot operation.
        
        Immediately stops current operations and enters STOPPED state.
        Robot must be reset before new operations can begin.
        
        Raises:
            APIError: If robot cannot be stopped
        """
        return await self._post("/api/v1/action/robot/status/stop")

    async def reset(self) -> None:
        """Reset robot to ready state.
        
        Clears error conditions and prepares robot for new operations.
        This does not perform physical homing - use initialize() for stable restart.
        
        Raises:
            APIError: If robot cannot be reset
        """
        return await self._post("/api/v1/action/robot/status/reset")


from typing import Callable, Coroutine

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.common.alma.enums import LocationType
from ablelabs.neon.common.alma.structs import Location, FlowRate, Speed


class PipetteAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.upper_module_pipette_pick_up_tip)
    async def pick_up_tip(
        self,
        well: str,
        optimize: bool = True,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_drop_tip)
    async def drop_tip(
        self,
        location_type: LocationType,
        well: str,
        optimize: bool = True,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_move_to)
    async def move_to(
        self,
        location: Location,
        task_after_z_up: Callable[[], Coroutine] = None,
        z_speed: Speed = None,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_ready_plunger)
    async def ready_plunger(
        self,
        flow_rate: FlowRate = None,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_blow_out)
    async def blow_out(
        self,
        flow_rate: FlowRate = None,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_aspirate)
    async def aspirate(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        flow_rate: float = None,
        height_offset: float = 0,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_dispense)
    async def dispense(
        self,
        volume: float,
        flow_rate: float = None,
        height_offset: float = 0,
    ):
        pass

    @run_server_func(RobotRouter.upper_module_pipette_mix)
    async def mix(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        iteration: int,
        flow_rate: float = None,
        delay: float = 0.0,
        height_offset: float = 0,
    ):
        pass



from setuptools import setup, find_packages
import os


# README 파일 읽기
def read_readme():
    readme_path = os.path.join(os.path.dirname(__file__), "README.md")
    if os.path.exists(readme_path):
        with open(readme_path, "r", encoding="utf-8") as f:
            return f.read()
    return "ABLE Labs 로봇 제어 API 패키지"


setup(
    name="ablelabs",
    version="0.4.13",
    description="ABLE Labs API",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    author="sypark",
    author_email="sy.park@ablelabsinc.com",
    url="https://github.com/ABLE-Labs/ABLE-API",
    install_requires=[
        "et-xmlfile>=1.1.0",
        "future>=1.0.0",
        "iso8601>=2.1.0",
        "loguru>=0.7.2",
        "openpyxl>=3.1.5",
        "pyserial>=3.5",
        "PyYAML>=6.0.1",
        "httpx>=0.28.1",
    ],
    packages=find_packages(),
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3.10",
    ],
)



from enum import Enum


class LEDColor(str, Enum):
    """LED Bar 색상 정의

    Available colors for LED bar control.
    """

    NONE = "NONE"
    RED = "RED"
    GREEN = "GREEN"
    BLUE = "BLUE"
    WHITE = "WHITE"
    CYAN = "CYAN"
    MAGENTA = "MAGENTA"
    YELLOW = "YELLOW"
    ORANGE = "ORANGE"


class ZReference(str, Enum):
    """Z-axis position reference point

    랩웨어 기준 높이:
    - BOTTOM_JUST: 랩웨어 바닥 표면
    - BOTTOM: 표준 액체 핸들링 높이 (bottom_offset 적용)
    - TOP_JUST: 랩웨어 웰 상단 표면
    - TOP: 안전 이동 높이 (top_offset 적용)

    팁 랙에서의 사용:
    - TOP_JUST: 팁 장착 시작 높이 (팁 상단)
    - BOTTOM_JUST: 팁 장착 완료 높이 (팁 홀더 바닥)
    - BOTTOM: 팁 탈착 높이 (cone_length만큼 덜 들어감)
    - TOP: 안전 이동 높이
    """

    BOTTOM_JUST = "BOTTOM_JUST"  # 랩웨어 바닥 표면
    BOTTOM = "BOTTOM"  # 바닥에서 offset 적용된 높이
    TOP_JUST = "TOP_JUST"  # 랩웨어 상단 표면
    TOP = "TOP"  # 상단에서 offset 적용된 안전 높이


class Axis(str, Enum):
    """Robot axis definitions
    
    Available axes for motion control:
    - X, Y: Horizontal movement axes
    - Z1, Z2: Vertical axes for pipette mounts 1 and 2
    - P1, P2: Plunger axes for pipettes 1 and 2
    """
    X = "x"
    Y = "y"
    Z1 = "z1"
    Z2 = "z2"
    P1 = "p1"
    P2 = "p2"


class PipetteType(str, Enum):
    """Pipette type definitions"""
    SINGLE = "single"
    MULTI = "multi"


class LabwareType(str, Enum):
    """Labware type definitions for configuration"""
    PLATE = "plate"
    RESERVOIR = "reservoir"
    TIP_RACK = "tip_rack"



import httpx
from .decorators import extract_data_async

class Base:
    def __init__(self, base_url: str):
        self._timeout = 300
        self._base_url = base_url.rstrip("/")
        self._client = httpx.AsyncClient(
            headers={"Content-Type": "application/json"},
            timeout=httpx.Timeout(self._timeout),
            limits=httpx.Limits(
                max_keepalive_connections=9999,
                max_connections=10,
            ),
        )

    @extract_data_async
    async def _get(self, path: str, params: dict = None) -> dict:
        url = f"{self._base_url}{path}"

        response = await self._client.get(url, params=params, timeout=self._timeout)

        # Check status code
        if response.status_code != 200:
            raise Exception(
                f"API returned error status {response.status_code}: {response.reason_phrase}"
            )

        return response.json() if response.text else {}

    @extract_data_async
    async def _post(
        self, path: str, body: dict = None, params: dict = None
    ):
        url = f"{self._base_url}{path}"

        response = await self._client.post(
            url, json=body, params=params, timeout=self._timeout
        )

        # Check status code
        if response.status_code != 200:
            raise Exception(
                f"API returned error status {response.status_code}: {response.reason_phrase}"
            )

        return response.json() if response.text else {}
